# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
import pytest

from libqtile.scripts.migrations import MIGRATIONS, load_migrations
from test.test_check import have_mypy, is_cpython

pytestmark = pytest.mark.skipif(not is_cpython() or not have_mypy(), reason="needs mypy")

migration_tests = []

# We store a list of test IDs so that tests can be filtered during development
# e.g. pytest -k MigrationID
migration_ids = []

load_migrations()

for m in MIGRATIONS:
    tests = []
    for i, test in enumerate(m.TESTS):
        if not test.check:
            continue
        tests.append((m.ID, test))
        migration_ids.append(f"{m.ID}-{i}")

    if not tests:
        tests.append((m.ID, None))
        migration_ids.append(f"{m.ID}-no-check-test")

    migration_tests.extend(tests)


@pytest.mark.parametrize("migration_tester", migration_tests, indirect=True, ids=migration_ids)
def test_check_all_migrations(migration_tester):
    migration_tester.assert_check()



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys

from setuptools import build_meta as _orig
from setuptools.build_meta import *  # noqa: F401,F403

WAYLAND_DEPENDENCIES = ["pywlroots>=0.17.0,<0.18.0"]


def wants_wayland(config_settings):
    if config_settings:
        for key in ["Backend", "backend"]:
            if config_settings.get(key, "").lower() == "wayland":
                return True

    return False


def get_requires_for_build_wheel(config_settings=None):
    """Inject pywlroots into build dependencies if wayland requested."""
    reqs = _orig.get_requires_for_build_wheel(config_settings)
    if wants_wayland(config_settings):
        reqs += WAYLAND_DEPENDENCIES

    return reqs


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Stop building if wayland requested but pywlroots is not installed."""
    if wants_wayland(config_settings):
        try:
            import wlroots  # noqa: F401
        except ImportError:
            sys.exit("Wayland backend requested but pywlroots is not installed.")

    return _orig.build_wheel(wheel_directory, config_settings, metadata_directory)



# -*- coding: utf-8 -*-
# Copyright (C) 2015, zordsdavini
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from functools import partial

from libqtile.widget import base


class Moc(base.ThreadPoolText):
    """A simple MOC widget.

    Show the artist and album of now listening song and allow basic mouse
    control from the bar:

    - toggle pause (or play if stopped) on left click;
    - skip forward in playlist on scroll up;
    - skip backward in playlist on scroll down.

    MOC (http://moc.daper.net) should be installed.
    """

    defaults = [
        ("play_color", "00ff00", "Text colour when playing."),
        ("noplay_color", "cecece", "Text colour when not playing."),
        ("update_interval", 0.5, "Update Time in seconds."),
    ]

    def __init__(self, **config):
        base.ThreadPoolText.__init__(self, "", **config)
        self.add_defaults(Moc.defaults)
        self.status = ""
        self.local = None

        self.add_callbacks(
            {
                "Button1": self.play,
                "Button4": partial(subprocess.Popen, ["mocp", "-f"]),
                "Button5": partial(subprocess.Popen, ["mocp", "-r"]),
            }
        )

    def get_info(self):
        """Return a dictionary with info about the current MOC status."""
        try:
            output = self.call_process(["mocp", "-i"])
        except subprocess.CalledProcessError as err:
            output = err.output
        if output.startswith("State"):
            output = output.splitlines()
            info = {"State": "", "File": "", "SongTitle": "", "Artist": "", "Album": ""}

            for line in output:
                for data in info:
                    if data in line:
                        info[data] = line[len(data) + 2 :].strip()
                        break
            return info

    def now_playing(self):
        """Return a string with the now playing info (Artist - Song Title)."""
        info = self.get_info()
        now_playing = ""
        if info:
            status = info["State"]
            if self.status != status:
                self.status = status
                if self.status == "PLAY":
                    self.layout.colour = self.play_color
                else:
                    self.layout.colour = self.noplay_color
            title = info["SongTitle"]
            artist = info["Artist"]
            if title and artist:
                now_playing = "♫ {0} - {1}".format(artist, title)
            elif title:
                now_playing = "♫ {0}".format(title)
            else:
                basename = os.path.basename(info["File"])
                filename = os.path.splitext(basename)[0]
                now_playing = "♫ {0}".format(filename)

            if self.status == "STOP":
                now_playing = "♫"

        return now_playing

    def play(self):
        """Play music if stopped, else toggle pause."""
        if self.status in ("PLAY", "PAUSE"):
            subprocess.Popen(["mocp", "-G"])
        elif self.status == "STOP":
            subprocess.Popen(["mocp", "-p"])

    def poll(self):
        """Poll content for the text box."""
        return self.now_playing()



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import sys
from importlib import reload
from types import ModuleType

import pytest

from libqtile.widget import df
from test.widgets.conftest import FakeBar


class FakeOS(ModuleType):
    class statvfs:  # noqa: N801
        def __init__(self, *args, **kwargs):
            pass

        @property
        def f_frsize(self):
            return 4096

        @property
        def f_blocks(self):
            return 60000000

        @property
        def f_bfree(self):
            return 15000000

        @property
        def f_bavail(self):
            return 10000000


# Patches os.stavfs gives these values for df widget:
#  unit: G
#  size = 228
#  free = 57
#  user_free = 38
#  ratio (user_free / size) = 83.3333%
@pytest.fixture()
def patched_df(monkeypatch):
    monkeypatch.setitem(sys.modules, "os", FakeOS("os"))
    reload(df)


@pytest.mark.usefixtures("patched_df")
def test_df_no_warning(fake_qtile, fake_window):
    """Test no text when free space over threshold"""
    df1 = df.DF()
    fakebar = FakeBar([df1], window=fake_window)
    df1._configure(fake_qtile, fakebar)
    text = df1.poll()
    assert text == ""

    df1.draw()
    assert df1.layout.colour == df1.foreground


@pytest.mark.usefixtures("patched_df")
def test_df_always_visible(fake_qtile, fake_window):
    """Test text is always displayed"""
    df2 = df.DF(visible_on_warn=False)
    fakebar = FakeBar([df2], window=fake_window)
    df2._configure(fake_qtile, fakebar)
    text = df2.poll()

    # See values above
    assert text == "/ (38G|83%)"

    df2.draw()
    assert df2.layout.colour == df2.foreground


@pytest.mark.usefixtures("patched_df")
def test_df_warn_space(fake_qtile, fake_window):
    """
    Test text is visible and colour changes when space
    below threshold
    """
    df3 = df.DF(warn_space=40)
    fakebar = FakeBar([df3], window=fake_window)
    df3._configure(fake_qtile, fakebar)
    text = df3.poll()

    # See values above
    assert text == "/ (38G|83%)"

    df3.draw()
    assert df3.layout.colour == df3.warn_color



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class UpdateTogroupTransformer(MigrationTransformer):
    @m.call_if_inside(
        m.Call(func=m.Name(m.MatchIfTrue(lambda n: "togroup" in n)))
        | m.Call(func=m.Attribute(attr=m.Name(m.MatchIfTrue(lambda n: "togroup" in n))))
    )
    @m.leave(m.Arg(keyword=m.Name("groupName")))
    def update_togroup_args(self, original_node, updated_node) -> cst.Arg:
        """Changes 'groupName' kwarg to 'group_name'."""
        self.lint(
            original_node, "The 'groupName' keyword argument should be replaced with 'group_name."
        )
        return updated_node.with_changes(keyword=cst.Name("group_name"))


class UpdateTogroupArgs(_QtileMigrator):
    ID = "UpdateTogroupArgs"

    SUMMARY = "Updates ``groupName`` keyword argument to ``group_name``."

    HELP = """
    To be consistent with codestyle, the ``groupName`` argument in the ``togroup`` command needs to be
    changed to ``group_name``.


    The following code:

    .. code:: python

        lazy.window.togroup(groupName="1")

    will result in a warning in your logfile: ``Window.togroup's groupName is deprecated; use group_name``.

    The code should be updated to:

    .. code:: python

        lazy.window.togroup(group_name="1")

    """

    AFTER_VERSION = "0.18.1"

    TESTS = [
        Check(
            """
            from libqtile.config import Key
            from libqtile.lazy import lazy

            k = Key([], 's', lazy.window.togroup(groupName="g"))
            c = lambda win: win.togroup(groupName="g")
            """,
            """
            from libqtile.config import Key
            from libqtile.lazy import lazy

            k = Key([], 's', lazy.window.togroup(group_name="g"))
            c = lambda win: win.togroup(group_name="g")
            """,
        )
    ]

    visitor = UpdateTogroupTransformer


add_migration(UpdateTogroupArgs)



# Copyright (c) 2021 Matt Colligan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pywayland.server import Listener
from wlroots.wlr_types import Output as WlrOutput
from wlroots.wlr_types import SceneTree
from wlroots.wlr_types.layer_shell_v1 import LayerShellV1Layer, LayerSurfaceV1

from libqtile import hook
from libqtile.backend.wayland.output import Output
from libqtile.backend.wayland.window import Static
from libqtile.command.base import expose_command
from libqtile.log_utils import logger

try:
    # Continue if ffi not built, so that docs can be built without wayland deps.
    from libqtile.backend.wayland._ffi import ffi
except ModuleNotFoundError:
    pass

if TYPE_CHECKING:
    from typing import Any

    from wlroots.wlr_types.scene import SceneLayerSurfaceV1

    from libqtile.backend.wayland.core import Core
    from libqtile.core.manager import Qtile
    from libqtile.utils import ColorsType


class LayerStatic(Static[LayerSurfaceV1]):
    """A static window belonging to the layer shell."""

    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        surface: LayerSurfaceV1,
        wid: int,
    ):
        Static.__init__(self, core, qtile, surface, wid)

        self.desired_width = 0
        self.desired_height = 0

        self.data_handle = ffi.new_handle(self)
        surface.data = self.data_handle

        # Determine which output this window is to appear on
        if wlr_output := surface.output:
            logger.debug("Layer surface requested output: %s", wlr_output.name)
        else:
            wlr_output = cast(
                WlrOutput, core.output_layout.output_at(core.cursor.x, core.cursor.y)
            )
            logger.debug("Layer surface given output: %s", wlr_output.name)
            surface.output = wlr_output

        output = cast(Output, wlr_output.data)
        self.output = output
        self.screen = output.screen

        # Add the window to the scene graph
        parent_tree = core.layer_trees[surface.pending.layer]
        self.scene_layer: SceneLayerSurfaceV1 = core.scene.layer_surface_v1_create(
            parent_tree, surface
        )
        self.tree: SceneTree = self.scene_layer.tree
        self.tree.node.data = self.data_handle
        self.popup_tree = SceneTree.create(parent_tree)  # Popups get their own tree
        self.popup_tree.node.data = self.data_handle

        # Set up listeners
        self.add_listener(surface.surface.map_event, self._on_map)
        self.add_listener(surface.surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.surface.commit_event, self._on_commit)

        # Temporarily set the layer's current state to pending so that we can easily
        # arrange it. TODO: how much of this is needed?
        self._layer = surface.pending.layer
        old_state = surface.current
        surface.current = surface.pending
        self.unhide()
        self.output.organise_layers()
        surface.current = old_state
        self._move_to_layer(old_state.layer)

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        if self.surface.output and self.surface.output.data:
            output = self.surface.output.data
            if output != self.output:
                # The window wants to move to a different output.
                if self.tree.node.enabled:
                    self.output.layers[self._layer].remove(self)
                    output.layers[self._layer].append(self)
                self.output = output

        pending = self.surface.pending
        if (
            self._layer != pending.layer
            or self._width != pending.desired_width
            or self._height != pending.desired_height
        ):
            # The window has changed its desired layer or dimensions.
            self._move_to_layer(pending.layer)

    def _move_to_layer(self, layer: LayerShellV1Layer) -> None:
        new_parent = self.core.layer_trees[layer]
        self.tree.node.reparent(new_parent)
        self.popup_tree.node.reparent(new_parent)

        if self.tree.node.enabled:
            # If we're mapped, we also need to update the lists on the output.
            self.output.layers[self._layer].remove(self)
            self.output.layers[layer].append(self)
            self.output.organise_layers()

        self._layer = layer

    def finalize(self) -> None:
        super().finalize()
        self.popup_tree.node.destroy()

    def kill(self) -> None:
        self.surface.destroy()

    def hide(self) -> None:
        if self.core.exclusive_layer is self:
            self.core.exclusive_layer = None

        if self.reserved_space:
            self.qtile.free_reserved_space(self.reserved_space, self.screen)
            self.reserved_space = None

        if self.surface.surface == self.core.seat.keyboard_state.focused_surface:
            group = self.qtile.current_screen.group
            if group.current_window:
                group.focus(group.current_window, warp=self.qtile.config.cursor_warp)
            else:
                self.core.seat.keyboard_clear_focus()

        if self in self.output.layers[self._layer]:
            self.tree.node.set_enabled(enabled=False)
            # TODO also toggle popup_tree
            self.output.layers[self._layer].remove(self)
            self.output.organise_layers()

    def unhide(self) -> None:
        if self not in self.output.layers[self._layer]:
            self.tree.node.set_enabled(enabled=True)
            self.output.layers[self._layer].append(self)
            self.output.organise_layers()

    def focus(self, _warp: bool = True) -> None:
        self.core.focus_window(self)
        hook.fire("client_focus", self)

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self.tree.node.set_position(x, y)
        self.popup_tree.node.set_position(x, y)
        # The actual resizing is done by `Output`.
        self._width = width
        self._height = height

    @expose_command()
    def bring_to_front(self) -> None:
        self.tree.node.raise_to_top()



# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import libqtile.config
import libqtile.confreader
import libqtile.layout
import libqtile.resources.default_config
from libqtile.widget import ScreenSplit


@pytest.fixture
def widget():
    yield ScreenSplit


# We need to override default minimal_conf so we can force the layout
@pytest.fixture(scope="function")
def minimal_conf_noscreen():
    class MinimalConf(libqtile.confreader.Config):
        auto_fullscreen = False
        keys = []
        mouse = []
        groups = [libqtile.config.Group("a"), libqtile.config.Group("b")]
        layouts = [libqtile.layout.ScreenSplit()]
        floating_layout = libqtile.resources.default_config.floating_layout
        screens = []

    return MinimalConf


def ss_screensplit(screenshot_manager):
    screenshot_manager.take_screenshot()



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

# The test_widget_init_config will cover the scenario where `kbdd` is not
# running and will also run the asyncio _config_async method.

# This test file covers the remaining widget code

from importlib import reload

import pytest

from test.widgets.conftest import FakeBar


async def mock_signal_receiver(*args, **kwargs):
    return True


class MockSpawn:
    call_count = 0

    @classmethod
    def call_process(cls, *args, **kwargs):
        if cls.call_count == 0:
            cls.call_count += 1
            return ""
        return "kbdd"


class MockMessage:
    def __init__(self, is_signal=True, body=0):
        self.message_type = 1 if is_signal else 0
        self.body = [body]


@pytest.fixture
def patched_widget(monkeypatch):
    from libqtile.widget import keyboardkbdd

    reload(keyboardkbdd)

    # The next line shouldn't be necessary but I got occasional failures without it when testing locally
    monkeypatch.setattr(
        "libqtile.widget.keyboardkbdd.KeyboardKbdd.call_process", MockSpawn.call_process
    )
    monkeypatch.setattr("libqtile.widget.keyboardkbdd.add_signal_receiver", mock_signal_receiver)
    return keyboardkbdd


def test_keyboardkbdd_process_running(fake_qtile, patched_widget, fake_window):
    MockSpawn.call_count = 1
    kbd = patched_widget.KeyboardKbdd(configured_keyboards=["gb", "us"])
    fakebar = FakeBar([kbd], window=fake_window)
    kbd._configure(fake_qtile, fakebar)
    assert kbd.is_kbdd_running
    assert kbd.keyboard == "gb"

    # Create a message with the index of the active keyboard
    message = MockMessage(body=1)
    kbd._signal_received(message)
    assert kbd.keyboard == "us"


def test_keyboardkbdd_process_not_running(fake_qtile, patched_widget, fake_window):
    MockSpawn.call_count = 0
    kbd = patched_widget.KeyboardKbdd(configured_keyboards=["gb", "us"])
    fakebar = FakeBar([kbd], window=fake_window)
    kbd._configure(fake_qtile, fakebar)
    assert not kbd.is_kbdd_running
    assert kbd.keyboard == "N/A"

    # Second call of _check_kbdd will confirm process running
    # so widget should now show layout
    kbd.poll()
    assert kbd.keyboard == "gb"


# Custom colours are not set until a signal is received
# TO DO: This should be fixed so the colour is set on __init__
def test_keyboard_kbdd_colours(fake_qtile, patched_widget, fake_window):
    MockSpawn.call_count = 1
    kbd = patched_widget.KeyboardKbdd(
        configured_keyboards=["gb", "us"], colours=["#ff0000", "#00ff00"]
    )
    fakebar = FakeBar([kbd], window=fake_window)
    kbd._configure(fake_qtile, fakebar)

    # Create a message with the index of the active keyboard
    message = MockMessage(body=0)
    kbd._signal_received(message)
    assert kbd.layout.colour == "#ff0000"

    # Create a message with the index of the active keyboard
    message = MockMessage(body=1)
    kbd._signal_received(message)
    assert kbd.layout.colour == "#00ff00"

    # No change where self.colours is a string
    kbd.colours = "#ffff00"
    kbd._set_colour(1)
    assert kbd.layout.colour == "#00ff00"

    # Colours list is shorter than length of layouts
    kbd.colours = ["#ff00ff"]

    # Should pick second item in colours list but it doesn't exit
    # so widget looks for previous item
    kbd._set_colour(1)
    assert kbd.layout.colour == "#ff00ff"



# Copyright (c) 2024 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from libqtile import layout
from libqtile.config import Bar, Screen
from libqtile.confreader import Config
from libqtile.widget import plasma


@pytest.fixture(scope="function")
def plasma_manager(manager_nospawn, request):
    class PlasmaConfig(Config):
        layouts = [layout.Plasma()]
        screens = [Screen(top=Bar([plasma.Plasma(**getattr(request, "param", dict()))], 30))]

    manager_nospawn.start(PlasmaConfig)

    yield manager_nospawn


def config(**kwargs):
    return pytest.mark.parametrize("plasma_manager", [kwargs], indirect=True)


def test_plasma_defaults(plasma_manager):
    def text():
        return plasma_manager.c.widget["plasma"].info()["text"]

    layout = plasma_manager.c.layout

    assert text() == " H"

    layout.mode_vertical()
    assert text() == " V"

    layout.mode_horizontal_split()
    assert text() == "HS"

    layout.mode_vertical_split()
    assert text() == "VS"


@config(horizontal="-", vertical="|", horizontal_split="=", vertical_split="||", format="{mode}")
def test_custom_text(plasma_manager):
    def text():
        return plasma_manager.c.widget["plasma"].info()["text"]

    layout = plasma_manager.c.layout

    assert text() == "-"

    layout.mode_vertical()
    assert text() == "|"

    layout.mode_horizontal_split()
    assert text() == "="

    layout.mode_vertical_split()
    assert text() == "||"


@config(format="{mode}")
def test_window_focus_change(plasma_manager):
    def text():
        return plasma_manager.c.widget["plasma"].info()["text"]

    def win(name):
        idx = [w["id"] for w in plasma_manager.c.windows() if w["name"] == name]
        assert idx
        return plasma_manager.c.window[idx[0]]

    layout = plasma_manager.c.layout

    assert text() == "H"
    plasma_manager.test_window("one")
    plasma_manager.test_window("two")

    layout.mode_vertical()
    plasma_manager.test_window("three")
    assert text() == "V"

    win("one").focus()
    assert text() == "H"

    win("three").focus()
    assert text() == "V"

    win("two").focus()
    assert text() == "V"

    win("one").focus()
    assert text() == "H"


@config(format="{mode}")
def test_mode_change(plasma_manager):
    def text():
        return plasma_manager.c.widget["plasma"].info()["text"]

    assert text() == "H"

    for mode in ["HS", "V", "VS", "H"]:
        plasma_manager.c.widget["plasma"].next_mode()
        assert text() == mode

    for mode in ["HS", "V", "VS", "H"]:
        plasma_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 1)
        assert text() == mode



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2013 Mattias Svala
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2014 Chris Wesseling
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_dimensions, assert_focus_path, assert_focused


class ZoomyConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
    ]
    layouts = [
        layout.Zoomy(columnwidth=200),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []


zoomy_config = pytest.mark.parametrize("manager", [ZoomyConfig], indirect=True)


@zoomy_config
def test_zoomy_one(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 600, 600)
    manager.test_window("two")
    assert_dimensions(manager, 0, 0, 600, 600)
    manager.test_window("three")
    assert_dimensions(manager, 0, 0, 600, 600)
    assert_focus_path(manager, "two", "one", "three")
    # TODO(pc) find a way to check size of inactive windows


@zoomy_config
def test_zoomy_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions, Zoomy adds clients at head
    assert manager.c.layout.info()["clients"] == ["three", "two", "one"]
    # last added window has focus
    assert_focused(manager, "three")

    # assert window focus cycle, according to order in layout
    assert_focus_path(manager, "two", "one", "float1", "float2", "three")



STATUS_NOTIFIER_ITEM_SPEC = """
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
  <interface name='org.kde.StatusNotifierItem'>
    <annotation name="org.gtk.GDBus.C.Name" value="Item" />
    <method name='ContextMenu'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='Activate'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='SecondaryActivate'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='Scroll'>
      <arg type='i' direction='in' name='delta'/>
      <arg type='s' direction='in' name='orientation'/>
    </method>
    <signal name='NewTitle'/>
    <signal name='NewIcon'/>
    <signal name='NewAttentionIcon'/>
    <signal name='NewOverlayIcon'/>
    <signal name='NewToolTip'/>
    <signal name='NewStatus'>
      <arg type='s' name='status'/>
    </signal>
    <property name='Category' type='s' access='read'/>
    <property name='Id' type='s' access='read'/>
    <property name='Title' type='s' access='read'/>
    <property name='Status' type='s' access='read'/>
    <!-- See discussion on pull #536
    <property name='WindowId' type='u' access='read'/>
    -->
    <property name='IconThemePath' type='s' access='read'/>
    <property name='IconName' type='s' access='read'/>
    <property name='IconPixmap' type='a(iiay)' access='read'/>
    <property name='OverlayIconName' type='s' access='read'/>
    <property name='OverlayIconPixmap' type='a(iiay)' access='read'/>
    <property name='AttentionIconName' type='s' access='read'/>
    <property name='AttentionIconPixmap' type='a(iiay)' access='read'/>
    <property name='AttentionMovieName' type='s' access='read'/>
    <property name='ToolTip' type='(sa(iiay)ss)' access='read'/>
    <property name='Menu' type='o' access='read'/>
    <property name='ItemIsMenu' type='b' access='read'/>
  </interface>
</node>
"""



# Copyright (c) 2019 Guangwang Huang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_focus_path, assert_focused


class TreeTabConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    layouts = [
        layout.TreeTab(sections=["Foo", "Bar"]),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


treetab_config = pytest.mark.parametrize("manager", [TreeTabConfig], indirect=True)


@treetab_config
def test_window(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1", floating=True)
    manager.test_window("float2", floating=True)
    manager.test_window("three")

    # test preconditions, columns adds clients at pos of current, in two stacks
    assert manager.c.layout.info()["clients"] == ["one", "three", "two"]
    assert manager.c.layout.info()["sections"] == ["Foo", "Bar"]
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two"], ["three"]],
        "Bar": [],
    }

    # last added window has focus
    assert_focused(manager, "three")
    manager.c.layout.up()
    assert_focused(manager, "two")
    manager.c.layout.down()
    assert_focused(manager, "three")

    # test command move_up/down
    manager.c.layout.move_up()
    assert manager.c.layout.info()["clients"] == ["one", "three", "two"]
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["three"], ["two"]],
        "Bar": [],
    }
    manager.c.layout.move_down()
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two"], ["three"]],
        "Bar": [],
    }

    # section_down/up
    manager.c.layout.up()  # focus two
    manager.c.layout.section_down()
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["three"]],
        "Bar": [["two"]],
    }
    manager.c.layout.section_up()
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["three"], ["two"]],
        "Bar": [],
    }

    # del_section
    manager.c.layout.up()  # focus three
    manager.c.layout.section_down()
    manager.c.layout.del_section("Bar")
    assert manager.c.layout.info()["client_trees"] == {"Foo": [["one"], ["two"], ["three"]]}

    # add_section
    manager.c.layout.add_section("Baz")
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two"], ["three"]],
        "Baz": [],
    }
    manager.c.layout.del_section("Baz")

    # move_left/right
    manager.c.layout.move_left()  # no effect for top-level children
    assert manager.c.layout.info()["client_trees"] == {"Foo": [["one"], ["two"], ["three"]]}
    manager.c.layout.move_right()
    assert manager.c.layout.info()["client_trees"] == {"Foo": [["one"], ["two", ["three"]]]}
    manager.c.layout.move_right()  # no effect
    assert manager.c.layout.info()["client_trees"] == {"Foo": [["one"], ["two", ["three"]]]}
    manager.test_window("four")
    manager.c.layout.move_right()
    manager.c.layout.up()
    manager.test_window("five")
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two", ["three", ["four"]], ["five"]]]
    }

    # expand/collapse_branch, and check focus order
    manager.c.layout.up()
    manager.c.layout.up()  # focus three
    manager.c.layout.collapse_branch()
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two", ["three"], ["five"]]]
    }
    assert_focus_path(manager, "five", "float1", "float2", "one", "two", "three")
    manager.c.layout.expand_branch()
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two", ["three", ["four"]], ["five"]]]
    }
    assert_focus_path(manager, "four", "five", "float1", "float2", "one", "two", "three")


@treetab_config
def test_sort_windows(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("101")
    manager.test_window("102")
    manager.test_window("103")
    assert manager.c.layout.info()["client_trees"] == {
        "Foo": [["one"], ["two"], ["101"], ["102"], ["103"]],
        "Bar": [],
    }
    """
    # TODO how to serialize a function object? i.e. `sorter`:

    def sorter(window):
        try:
            if int(window.name) % 2 == 0:
                return 'Even'
            else:
                return 'Odd'
        except ValueError:
            return 'Bar'

    manager.c.layout.sort_windows(sorter)
    assert manager.c.layout.info()['client_trees'] == {
        'Foo': [],
        'Bar': [['one'], ['two']],
        'Even': [['102']],
        'Odd': [['101'], ['103']]
    }
    """



# Copyright (c) 2024 Florian G. Hechler

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import sys
import tempfile
from importlib import reload
from types import ModuleType

import pytest

import libqtile.config
import libqtile.widget
from libqtile.bar import Bar


class MockPsutil(ModuleType):
    pass


# Set ticks of fake stat file
def set_io_ticks(temp_file, io_ticks):
    temp_file.truncate(0)
    temp_file.write(
        f"13850 7547 1201386 14536 12612 60213 1838656 73847 0 {io_ticks} 112672 0 0 0 0 2116 24287"
    )
    temp_file.flush()


@pytest.fixture
def hdd_manager(monkeypatch, manager_nospawn, minimal_conf_noscreen):
    monkeypatch.setitem(sys.modules, "psutil", MockPsutil("psutil"))
    from libqtile.widget import hdd

    reload(hdd)

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=Bar([hdd.HDD()], 10))]

    manager_nospawn.start(config)
    yield manager_nospawn


def test_hdd(hdd_manager):
    # Create a fake stat file
    fake_stat_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)

    widget = hdd_manager.c.widget["hdd"]
    widget.eval(f"self.path = '{fake_stat_file.name}'")

    set_io_ticks(fake_stat_file, 0)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "HDD 0.0%"

    set_io_ticks(fake_stat_file, 300000)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "HDD 50.0%"

    set_io_ticks(fake_stat_file, 900000)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "HDD 100.0%"

    set_io_ticks(fake_stat_file, 2000000)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "HDD 100.0%"

    set_io_ticks(fake_stat_file, 0)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "HDD 0.0%"



# Copyright (c) 2021 Jeroen Wijenbergh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import libqtile
from libqtile import config
from libqtile.backend.x11.core import Core
from libqtile.confreader import Config
from libqtile.lazy import lazy


# Function that increments a counter
@lazy.function
def swallow_inc(qtile):
    qtile.test_data += 1
    return True


# Config with multiple keys and swallow parameters
class SwallowConfig(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        @libqtile.hook.subscribe.startup
        def _():
            libqtile.qtile.test_data = 0

    keys = [
        config.Key(
            ["control"],
            "k",
            swallow_inc(),
        ),
        config.Key(["control"], "j", swallow_inc(), swallow=False),
        config.Key(["control"], "i", swallow_inc().when(layout="idonotexist")),
        config.Key(
            ["control"],
            "o",
            swallow_inc().when(layout="idonotexist"),
            swallow_inc(),
        ),
    ]


# Helper to send process_key_event to the core manager
# It also looks up the keysym and mask to pass to it
def send_process_key_event(manager, key):
    keysym, mask = Core.lookup_key(None, key)
    output = manager.c.eval(f"self.process_key_event({keysym}, {mask})[1]")
    # Assert if eval successful
    assert output[0]
    # Convert the string to a bool
    return output[1] == "True"


def get_test_counter(manager):
    output = manager.c.eval("self.test_data")
    # Assert if eval successful
    assert output[0]
    return int(output[1])


@pytest.mark.parametrize("manager", [SwallowConfig], indirect=True)
def test_swallow(manager):
    # The first key needs to be True as swallowing is not set here
    # We expect the second key to not be handled, as swallow is set to False
    # The third needs to not be swallowed as the layout .when(...) check does not succeed
    # The fourth needs to be True as one of the functions is executed due to passing the .when(...) check
    expectedexecuted = [True, True, False, True]
    expectedswallow = [True, False, False, True]

    # Loop over all the keys in the config and assert
    prev_counter = 0
    for index, key in enumerate(SwallowConfig.keys):
        assert send_process_key_event(manager, key) == expectedswallow[index]

        # Test if the function was executed like we expected
        counter = get_test_counter(manager)
        if expectedexecuted[index]:
            assert counter > prev_counter
        else:
            assert counter == prev_counter
        prev_counter = counter

    not_used_key = config.Key(
        ["control"],
        "h",
        swallow_inc(),
    )

    # This key is not defined in the config so it should not be handled
    assert not send_process_key_event(manager, not_used_key)

    # This key is not defined so test data is not incremented
    assert get_test_counter(manager) == prev_counter



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from datetime import timedelta

import pytest

from libqtile.widget import pomodoro
from test.widgets.test_pomodoro import MockDatetime


def increment_time(self, increment):
    MockDatetime._adjustment += timedelta(minutes=increment)


@pytest.fixture
def widget(monkeypatch):
    monkeypatch.setattr("libqtile.widget.pomodoro.datetime", MockDatetime)
    pomodoro.Pomodoro.adjust_time = increment_time
    yield pomodoro.Pomodoro


def ss_pomodoro(screenshot_manager):
    bar = screenshot_manager.c.bar["top"]
    widget = screenshot_manager.c.widget["pomodoro"]

    # Inactive
    screenshot_manager.take_screenshot()

    bar.fake_button_press(0, "top", 0, 0, 3)
    widget.eval("self.update(self.poll())")

    # Active
    screenshot_manager.take_screenshot()

    widget.eval("self.adjust_time(25)")
    widget.eval("self.update(self.poll())")

    # Short break
    screenshot_manager.take_screenshot()



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class KeychordTransformer(MigrationTransformer):
    @m.leave(
        m.Call(
            func=m.Name("KeyChord"),
            args=[m.ZeroOrMore(), m.Arg(keyword=m.Name("mode")), m.ZeroOrMore()],
        )
    )
    def update_keychord_args(self, original_node, updated_node) -> cst.Call:
        """Changes 'mode' kwarg to 'mode' and 'value' kwargs."""
        args = original_node.args

        # (shouldn't be possible) if there are no args, stop here
        if not args:
            return original_node

        pos = 0

        # Look for the "mode" kwarg and get its position and value
        for i, arg in enumerate(args):
            if kwarg := arg.keyword:
                if kwarg.value == "mode":
                    # Ignore "mode" if it's already True or False
                    if m.matches(arg.value, (m.Name("True") | m.Name("False"))):
                        return original_node
                    pos = i
                    break
        # "mode" wasn't set so we can stop here
        else:
            return original_node

        self.lint(
            arg,  # We can pass the argument here which ensures the position is reported correctly
            "The use of mode='mode name' for KeyChord is deprecated. Use mode=True and value='mode name'.",
        )

        # Create two new kwargs
        # LibCST nodes are immutable so calling "with_changes" returns a new node
        name_arg = arg.with_changes(keyword=cst.Name("name"))
        mode_arg = arg.with_changes(value=cst.Name("True"))

        # Get the existing args and remove "mode"
        new_args = [a for i, a in enumerate(args) if i != pos]

        # Add "mode" and "value" kwargs
        new_args += [name_arg, mode_arg]

        # Return the updated node
        return updated_node.with_changes(args=new_args)


class KeychordArgs(_QtileMigrator):
    ID = "UpdateKeychordArgs"

    SUMMARY = "Updates ``KeyChord`` argument signature."

    HELP = """
    Previously, users could make a key chord persist by setting the `mode` to a string representing
    the name of the mode. For example:

    .. code:: python

        keys = [
            KeyChord(
                [mod],
                "x",
                [
                    Key([], "Up", lazy.layout.grow()),
                    Key([], "Down", lazy.layout.shrink())
                ],
                mode="Resize layout",
            )
        ]

    This will now result in the following warning message in the log file:

    .. code::

        The use of `mode` to set the KeyChord name is deprecated. Please use `name='Resize Layout'` instead.
        'mode' should be a boolean value to set whether the chord is persistent (True) or not."

    To remove the error, the config should be amended as follows:

    .. code:: python

        keys = [
            KeyChord(
                [mod],
                "x",
                [
                    Key([], "Up", lazy.layout.grow()),
                    Key([], "Down", lazy.layout.shrink())
                ],
                name="Resize layout",
            mode=True,
            )
        ]

    .. note::

       The formatting of the inserted argument may not correctly match your own formatting. You may this
       to run a tool like ``black`` after applying this migration to tidy up your code.

    """

    AFTER_VERSION = "0.21.0"

    TESTS = [
        Check(
            """
            from libqtile.config import Key, KeyChord
            from libqtile.lazy import lazy

            mod = "mod4"

            keys = [
                KeyChord(
                    [mod],
                    "x",
                    [
                        Key([], "Up", lazy.layout.grow()),
                        Key([], "Down", lazy.layout.shrink())
                    ],
                    mode="Resize layout",
                )
            ]
            """,
            """
            from libqtile.config import Key, KeyChord
            from libqtile.lazy import lazy

            mod = "mod4"

            keys = [
                KeyChord(
                    [mod],
                    "x",
                    [
                        Key([], "Up", lazy.layout.grow()),
                        Key([], "Down", lazy.layout.shrink())
                    ],
                    name="Resize layout",
                mode=True,
                )
            ]
            """,
        )
    ]

    visitor = KeychordTransformer


add_migration(KeychordArgs)



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from libqtile.scripts.migrations._base import (
    Check,
    RenamerTransformer,
    _QtileMigrator,
    add_migration,
)


class RenameThreadedPollTextTransformer(RenamerTransformer):
    from_to = ("ThreadedPollText", "ThreadPoolText")


class RenameThreadedPollText(_QtileMigrator):
    ID = "RenameThreadedPollText"

    SUMMARY = "Replaces ``ThreadedPollText`` with ``ThreadPoolText``."

    HELP = """
    The ``ThreadedPollText`` class needs to replced with ``ThreadPoolText``.

    This is because the ``ThreadPoolText`` class can do everything that the
    ``ThreadedPollText`` does so the redundant code was removed.

    Example:

    .. code:: python

        from libqtile import widget

        class MyPollingWidget(widget.base.ThreadedPollText):
            ...

    Should be updated as follows:

    .. code:: python

        from libqtile import widget

        class MyPollingWidget(widget.base.ThreadPoolText):
            ...

    """

    AFTER_VERSION = "0.16.1"

    TESTS = [
        Check(
            """
            from libqtile.widget.base import ThreadedPollText

            class MyWidget(ThreadedPollText):
                pass
            """,
            """
            from libqtile.widget.base import ThreadPoolText

            class MyWidget(ThreadPoolText):
                pass
            """,
        )
    ]

    visitor = RenameThreadedPollTextTransformer


add_migration(RenameThreadedPollText)



# Copyright (c) 2024 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from libqtile import config, widget
from libqtile.bar import Bar
from test.helpers import BareConfig


class GroupBoxConfig(BareConfig):
    screens = [
        config.Screen(
            top=Bar([widget.GroupBox(), widget.GroupBox(name="has_markup", markup=True)], 24)
        )
    ]
    groups = [config.Group("1", label="<sup>1</sup>")]


groupbox_config = pytest.mark.parametrize("manager", [GroupBoxConfig], indirect=True)


@groupbox_config
def test_groupbox_markup(manager):
    """Group labels can support markup but this is disabled by default."""
    no_markup = manager.c.widget["groupbox"]
    has_markup = manager.c.widget["has_markup"]

    # If markup is disabled, text will include markup tags so widget will be wider
    assert no_markup.info()["width"] > has_markup.info()["width"]



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import datetime
import sys
from importlib import reload

import pytest

import libqtile.config
from libqtile.widget import clock
from test.widgets.conftest import FakeBar


def no_op(*args, **kwargs):
    pass


# Mock Datetime object that returns a set datetime and also
# has a simplified timezone method to check functionality of
# the widget.
class MockDatetime(datetime.datetime):
    @classmethod
    def now(cls, *args, **kwargs):
        return cls(2021, 1, 1, 10, 20, 30)

    def astimezone(self, tzone=None):
        if tzone is None:
            return self
        return self + tzone.utcoffset(None)


@pytest.fixture
def patched_clock(monkeypatch):
    # Stop system importing these modules in case they exist on environment
    monkeypatch.setitem(sys.modules, "pytz", None)
    monkeypatch.setitem(sys.modules, "dateutil", None)
    monkeypatch.setitem(sys.modules, "dateutil.tz", None)

    # Reload module to force ImportErrors
    reload(clock)

    # Override datetime.
    # This is key for testing as we can fix time.
    monkeypatch.setattr("libqtile.widget.clock.datetime", MockDatetime)


def test_clock(fake_qtile, monkeypatch, fake_window):
    """test clock output with default settings"""
    monkeypatch.setattr("libqtile.widget.clock.datetime", MockDatetime)
    clk1 = clock.Clock()
    fakebar = FakeBar([clk1], window=fake_window)
    clk1._configure(fake_qtile, fakebar)
    text = clk1.poll()
    assert text == "10:20"


@pytest.mark.usefixtures("patched_clock")
def test_clock_invalid_timezone(fake_qtile, monkeypatch, fake_window, caplog):
    """test clock widget with invalid timezone (and no pytz or dateutil modules)"""

    class FakeDateutilTZ:
        @classmethod
        def tz(cls):
            return cls

        @classmethod
        def gettz(cls, val):
            return None

    # pytz and dateutil must not be in the sys.modules dict...
    monkeypatch.delitem(sys.modules, "pytz")
    monkeypatch.delitem(sys.modules, "dateutil")

    # Set up references to pytz and dateutil so we know these aren't being used
    # If they're called, the widget would try to run None(self.timezone) which
    # would raise an exception
    clock.pytz = None
    clock.dateutil = FakeDateutilTZ

    # Fake datetime module just adds the timezone value to the time
    clk2 = clock.Clock(timezone="1")

    fakebar = FakeBar([clk2], window=fake_window)
    clk2._configure(fake_qtile, fakebar)

    # An invalid timezone results in a log message
    assert (
        "Clock widget can not infer its timezone from a string without pytz or dateutil."
        in caplog.text
    )


@pytest.mark.usefixtures("patched_clock")
def test_clock_datetime_timezone(fake_qtile, monkeypatch, fake_window):
    """test clock with datetime timezone"""

    class FakeDateutilTZ:
        class TZ:
            @classmethod
            def gettz(cls, val):
                None

        tz = TZ

    # Set up references to pytz and dateutil so we know these aren't being used
    # If they're called, the widget would try to run None(self.timezone) which
    # would raise an exception
    clock.pytz = None
    clock.dateutil = FakeDateutilTZ

    # Fake datetime module just adds the timezone value to the time
    tz = datetime.timezone(datetime.timedelta(hours=1))
    clk3 = clock.Clock(timezone=tz)

    fakebar = FakeBar([clk3], window=fake_window)
    clk3._configure(fake_qtile, fakebar)
    text = clk3.poll()

    # Default time is 10:20 and we add 1 hour for the timezone
    assert text == "11:20"


@pytest.mark.usefixtures("patched_clock")
def test_clock_pytz_timezone(fake_qtile, monkeypatch, fake_window):
    """test clock with pytz timezone"""

    class FakeDateutilTZ:
        class TZ:
            @classmethod
            def gettz(cls, val):
                None

        tz = TZ

    class FakePytz:
        # pytz timezone is a string so convert it to an int and add 1
        # to show that this code is being run
        @classmethod
        def timezone(cls, value):
            hours = int(value) + 1
            return datetime.timezone(datetime.timedelta(hours=hours))

    # We need pytz in the sys.modules dict
    monkeypatch.setitem(sys.modules, "pytz", True)

    # Set up references to pytz and dateutil so we know these aren't being used
    # If they're called, the widget would try to run None(self.timezone) which
    # would raise an exception
    clock.pytz = FakePytz
    clock.dateutil = FakeDateutilTZ

    # Pytz timezone must be a string
    clk4 = clock.Clock(timezone="1")

    fakebar = FakeBar([clk4], window=fake_window)
    clk4._configure(fake_qtile, fakebar)
    text = clk4.poll()

    # Default time is 10:20 and we add 1 hour for the timezone plus and extra
    # 1 for the pytz function
    assert text == "12:20"


@pytest.mark.usefixtures("patched_clock")
def test_clock_dateutil_timezone(fake_qtile, monkeypatch, fake_window):
    """test clock with dateutil timezone"""

    class FakeDateutilTZ:
        class TZ:
            @classmethod
            def gettz(cls, val):
                hours = int(val) + 2
                return datetime.timezone(datetime.timedelta(hours=hours))

        tz = TZ

    # pytz must not be in the sys.modules dict...
    monkeypatch.delitem(sys.modules, "pytz")

    # ...but dateutil must be
    monkeypatch.setitem(sys.modules, "dateutil", True)

    # Set up references to pytz and dateutil so we know these aren't being used
    # If they're called, the widget would try to run None(self.timezone) which
    # would raise an exception
    clock.pytz = None
    clock.dateutil = FakeDateutilTZ

    # Pytz timezone must be a string
    clk5 = clock.Clock(timezone="1")

    fakebar = FakeBar([clk5], window=fake_window)
    clk5._configure(fake_qtile, fakebar)
    text = clk5.poll()

    # Default time is 10:20 and we add 1 hour for the timezone plus and extra
    # 1 for the pytz function
    assert text == "13:20"


@pytest.mark.usefixtures("patched_clock")
def test_clock_tick(manager_nospawn, minimal_conf_noscreen, monkeypatch):
    """Test clock ticks"""

    class FakeDateutilTZ:
        class TZ:
            @classmethod
            def gettz(cls, val):
                return int(val) + 2

        tz = TZ

    class TickingDateTime(datetime.datetime):
        offset = 0

        @classmethod
        def now(cls, *args, **kwargs):
            return cls(2021, 1, 1, 10, 20, 30)

        # This will return 10:20 on first call and 10:21 on all
        # subsequent calls
        def astimezone(self, tzone=None):
            extra = datetime.timedelta(minutes=TickingDateTime.offset)
            if TickingDateTime.offset < 1:
                TickingDateTime.offset += 1

            if tzone is None:
                return self + extra
            return self + datetime.timedelta(hours=tzone) + extra

    # pytz must not be in the sys.modules dict...
    monkeypatch.delitem(sys.modules, "pytz")

    # ...but dateutil must be
    monkeypatch.setitem(sys.modules, "dateutil", True)

    # Override datetime
    monkeypatch.setattr("libqtile.widget.clock.datetime", TickingDateTime)

    # Set up references to pytz and dateutil so we know these aren't being used
    # If they're called, the widget would try to run None(self.timezone) which
    # would raise an exception
    clock.pytz = None
    clock.dateutil = FakeDateutilTZ

    # set a long update interval as we'll tick manually
    clk6 = clock.Clock(update_interval=100)

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([clk6], 10))]

    manager_nospawn.start(config)

    topbar = manager_nospawn.c.bar["top"]
    manager_nospawn.c.widget["clock"].eval("self.tick()")
    assert topbar.info()["widgets"][0]["text"] == "10:21"


@pytest.mark.usefixtures("patched_clock")
def test_clock_change_timezones(fake_qtile, monkeypatch, fake_window):
    """test commands to change timezones"""

    tz1 = datetime.timezone(datetime.timedelta(hours=1))
    tz2 = datetime.timezone(-datetime.timedelta(hours=1))

    # Pytz timezone must be a string
    clk4 = clock.Clock(timezone=tz1)

    fakebar = FakeBar([clk4], window=fake_window)
    clk4._configure(fake_qtile, fakebar)
    text = clk4.poll()
    assert text == "11:20"

    clk4.update_timezone(tz2)
    text = clk4.poll()
    assert text == "09:20"

    clk4.use_system_timezone()
    text = clk4.poll()
    assert text == "10:20"



#!/usr/bin/env python3
"""
This creates a minimal window using GTK that works the same in both X11 or Wayland.

GTK sets the window class via `--name <class>`, and then we manually set the window
title and type. Therefore this is intended to be called as:

    python window.py --name <class> <title> <type>

where <type> is "normal" or "notification"

The window will close itself if it receives any key or button press events.
"""
# flake8: noqa

# This is needed otherwise the window will use any Wayland session it can find even if
# WAYLAND_DISPLAY is not set.
import os

if os.environ.get("WAYLAND_DISPLAY"):
    os.environ["GDK_BACKEND"] = "wayland"
else:
    os.environ["GDK_BACKEND"] = "x11"

# Disable GTK ATK bridge, which appears to trigger errors with e.g. test_strut_handling
# https://wiki.gnome.org/Accessibility/Documentation/GNOME2/Mechanics
os.environ["NO_AT_BRIDGE"] = "1"

import sys
from pathlib import Path

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from dbus_next import Message
from dbus_next.auth import Authenticator
from dbus_next.constants import MessageType, PropertyAccess
from dbus_next.glib.message_bus import MessageBus, _AuthLineSource
from dbus_next.service import ServiceInterface, dbus_property, method, signal
from gi.repository import Gdk, GLib, Gtk

ICON = Path(__file__).parent / "qtile_icon.rgba"


# This patch is needed to address the issue described here:
# https://github.com/altdesktop/python-dbus-next/issues/113
# Once dbus_next 0.2.4 is released, the patch can be removed.
class PatchedMessageBus(MessageBus):
    def _authenticate(self, authenticate_notify):
        self._stream.write(b"\0")
        first_line = self._auth._authentication_start()
        if first_line is not None:
            if type(first_line) is not str:
                raise AuthError("authenticator gave response not type str")
            self._stream.write(f"{first_line}\r\n".encode())
            self._stream.flush()

        def line_notify(line):
            try:
                resp = self._auth._receive_line(line)
                self._stream.write(Authenticator._format_line(resp))
                self._stream.flush()
                if resp == "BEGIN":
                    self._readline_source.destroy()
                    authenticate_notify(None)
                    return True
            except Exception as e:
                authenticate_notify(e)
                return True

        readline_source = _AuthLineSource(self._stream)
        readline_source.set_callback(line_notify)
        readline_source.add_unix_fd(self._fd, GLib.IO_IN)
        readline_source.attach(self._main_context)
        self._readline_source = readline_source


class SNItem(ServiceInterface):
    """
    Simplified StatusNotifierItem interface.

    Only exports methods, properties and signals required by
    StatusNotifier widget.
    """

    def __init__(self, window, *args):
        ServiceInterface.__init__(self, *args)
        self.window = window
        self.fullscreen = False

        with open(ICON, "rb") as f:
            self.icon = f.read()

        arr = bytearray(self.icon)
        for i in range(0, len(arr), 4):
            r, g, b, a = arr[i : i + 4]
            arr[i] = a
            arr[i + 1 : i + 4] = bytearray([r, g, b])

        self.icon = bytes(arr)

    @method()
    def Activate(self, x: "i", y: "i"):
        if self.fullscreen:
            self.window.unfullscreen()
        else:
            self.window.fullscreen()

        self.fullscreen = not self.fullscreen

    @dbus_property(PropertyAccess.READ)
    def IconName(self) -> "s":
        return ""

    @dbus_property(PropertyAccess.READ)
    def IconPixmap(self) -> "a(iiay)":
        return [[32, 32, self.icon]]

    @dbus_property(PropertyAccess.READ)
    def AttentionIconPixmap(self) -> "a(iiay)":
        return []

    @dbus_property(PropertyAccess.READ)
    def OverlayIconPixmap(self) -> "a(iiay)":
        return []

    @signal()
    def NewIcon(self):
        pass

    @signal()
    def NewAttentionIcon(self):
        pass

    @signal()
    def NewOverlayIcon(self):
        pass


if __name__ == "__main__":
    # GTK consumes the `--name <class>` args
    if len(sys.argv) > 1:
        title = sys.argv[1]
    else:
        title = "TestWindow"

    if len(sys.argv) > 2:
        window_type = sys.argv[2]
    else:
        window_type = "normal"

    # Check if we want to export a StatusNotifierItem interface
    sni = "export_sni_interface" in sys.argv

    win = Gtk.Window(title=title)
    win.set_default_size(100, 100)

    if window_type == "notification":
        if os.environ["GDK_BACKEND"] == "wayland":
            try:
                gi.require_version("GtkLayerShell", "0.1")
                from gi.repository import GtkLayerShell
            except ValueError:
                sys.exit(1)
            win.add(Gtk.Label(label="This is a test notification"))
            GtkLayerShell.init_for_window(win)

        else:
            win.set_type_hint(Gdk.WindowTypeHint.NOTIFICATION)

    elif window_type == "normal":
        win.set_type_hint(Gdk.WindowTypeHint.NORMAL)

    if sni:
        bus = PatchedMessageBus().connect_sync()

        item = SNItem(win, "org.kde.StatusNotifierItem")

        # Export interfaces on the bus
        bus.export("/StatusNotifierItem", item)

        # Request the service name
        bus.request_name_sync(f"test.qtile.window-{title.replace(' ','-')}")

        msg = bus.call_sync(
            Message(
                message_type=MessageType.METHOD_CALL,
                destination="org.freedesktop.StatusNotifierWatcher",
                interface="org.freedesktop.StatusNotifierWatcher",
                path="/StatusNotifierWatcher",
                member="RegisterStatusNotifierItem",
                signature="s",
                body=[bus.unique_name],
            )
        )

    win.connect("destroy", Gtk.main_quit)
    win.connect("key-press-event", Gtk.main_quit)
    win.show_all()

    Gtk.main()



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012 Tycho Andersen
# Copyright (c) 2014 Sean Vig
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from libqtile import config, ipc, resources
from libqtile.command.interface import IPCCommandInterface
from libqtile.confreader import Config
from libqtile.layout import Max
from libqtile.sh import QSh


class ShConfig(Config):
    keys = []
    mouse = []
    groups = [
        config.Group("a"),
        config.Group("b"),
    ]
    layouts = [
        Max(),
    ]
    floating_layout = resources.default_config.floating_layout
    screens = [config.Screen()]


sh_config = pytest.mark.parametrize("manager", [ShConfig], indirect=True)


@sh_config
def test_columnize(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh.columnize(["one", "two"]) == "one  two"

    sh.termwidth = 1
    assert sh.columnize(["one", "two"], update_termwidth=False) == "one\ntwo"

    sh.termwidth = 15
    v = sh.columnize(["one", "two", "three", "four", "five"], update_termwidth=False)
    assert v == "one    two  \nthree  four \nfive "


@sh_config
def test_ls(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh.do_ls(None) == "bar/     group/   layout/  screen/  widget/  window/  core/  "
    assert sh.do_ls("") == "bar/     group/   layout/  screen/  widget/  window/  core/  "
    assert sh.do_ls("layout") == "layout/group/   layout/window/  layout/screen/  layout[0]/    "

    assert sh.do_cd("layout") == "layout"
    assert sh.do_ls(None) == "group/   window/  screen/"
    assert (
        sh.do_ls("screen")
        == "screen/layout/  screen/window/  screen/bar/     screen/widget/  screen/group/ "
    )


@sh_config
def test_do_cd(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh.do_cd("layout") == "layout"
    assert sh.do_cd("../layout/0") == "layout[0]"
    assert sh.do_cd("..") == "/"
    assert sh.do_cd("layout") == "layout"
    assert sh.do_cd("../layout0/wibble") == "No such path."
    assert sh.do_cd(None) == "/"


@sh_config
def test_call(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh.process_line("status()") == "OK"

    v = sh.process_line("nonexistent()")
    assert v == "Command does not exist: nonexistent"

    v = sh.process_line("status(((")
    assert v == "Invalid command: status((("

    v = sh.process_line("status(1)")
    assert v.startswith("Caught command exception")


@sh_config
def test_complete(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh._complete("c", "c") == [
        "cd",
        "change_window_order",
        "commands",
        "critical",
    ]

    assert sh._complete("cd l", "l") == ["layout/"]
    assert sh._complete("cd layout/", "layout/") == [
        "layout/" + x for x in ["group", "window", "screen", "0"]
    ]
    assert sh._complete("cd layout/", "layout/g") == ["layout/group/"]


@sh_config
def test_help(manager):
    client = ipc.Client(manager.sockfile)
    command = IPCCommandInterface(client)
    sh = QSh(command)
    assert sh.do_help("nonexistent").startswith("No such command")
    assert sh.do_help("help")



# -*- coding: utf-8 -*-
# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2011 Kenji_Takahashi
# Copyright (c) 2012 roger
# Copyright (c) 2012, 2014 Tycho Andersen
# Copyright (c) 2012 Maximilian Köhl
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import itertools
import os

from libqtile import bar, hook
from libqtile.images import Img
from libqtile.log_utils import logger
from libqtile.widget import base


class CurrentLayout(base._TextBox):
    """
    Display the name of the current layout of the current group of the screen,
    the bar containing the widget, is on.
    """

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        layout_id = self.bar.screen.group.current_layout
        self.text = self.bar.screen.group.layouts[layout_id].name
        self.setup_hooks()

        self.add_callbacks(
            {
                "Button1": qtile.next_layout,
                "Button2": qtile.prev_layout,
            }
        )

    def hook_response(self, layout, group):
        if group.screen is not None and group.screen == self.bar.screen:
            self.text = layout.name
            self.bar.draw()

    def setup_hooks(self):
        hook.subscribe.layout_change(self.hook_response)

    def remove_hooks(self):
        hook.unsubscribe.layout_change(self.hook_response)

    def finalize(self):
        self.remove_hooks()
        base._TextBox.finalize(self)


class CurrentLayoutIcon(base._TextBox):
    """
    Display the icon representing the current layout of the
    current group of the screen on which the bar containing the widget is.

    If you are using custom layouts, a default icon with question mark
    will be displayed for them. If you want to use custom icon for your own
    layout, for example, `FooGrid`, then create a file named
    "layout-foogrid.png" and place it in `~/.icons` directory. You can as well
    use other directories, but then you need to specify those directories
    in `custom_icon_paths` argument for this plugin.

    The widget will look for icons with a `png` or `svg` extension.

    The order of icon search is:

    - dirs in `custom_icon_paths` config argument
    - `~/.icons`
    - built-in qtile icons
    """

    orientations = base.ORIENTATION_HORIZONTAL

    defaults = [
        ("scale", 1, "Scale factor relative to the bar height. Defaults to 1"),
        (
            "custom_icon_paths",
            [],
            "List of folders where to search icons before"
            "using built-in icons or icons in ~/.icons dir.  "
            "This can also be used to provide"
            "missing icons for custom layouts.  "
            "Defaults to empty list.",
        ),
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(CurrentLayoutIcon.defaults)

        self.length_type = bar.STATIC
        self.length = 0

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        layout_id = self.bar.screen.group.current_layout
        self.text = self.bar.screen.group.layouts[layout_id].name
        self.current_layout = self.text
        self.icons_loaded = False
        self.icon_paths = []
        self.surfaces = {}
        self._update_icon_paths()
        self._setup_images()
        self._setup_hooks()

        self.add_callbacks(
            {
                "Button1": qtile.next_layout,
                "Button2": qtile.prev_layout,
            }
        )

    def hook_response(self, layout, group):
        if group.screen is not None and group.screen == self.bar.screen:
            self.current_layout = layout.name
            self.bar.draw()

    def _setup_hooks(self):
        """
        Listens for layout change and performs a redraw when it occurs.
        """
        hook.subscribe.layout_change(self.hook_response)

    def _remove_hooks(self):
        """
        Listens for layout change and performs a redraw when it occurs.
        """
        hook.unsubscribe.layout_change(self.hook_response)

    def draw(self):
        if self.icons_loaded:
            try:
                surface = self.surfaces[self.current_layout]
            except KeyError:
                logger.error("No icon for layout %s", self.current_layout)
            else:
                self.drawer.clear(self.background or self.bar.background)
                self.drawer.ctx.save()
                self.drawer.ctx.translate(
                    (self.width - surface.width) / 2,
                    (self.bar.height - surface.height) / 2,
                )
                self.drawer.ctx.set_source(surface.pattern)
                self.drawer.ctx.paint()
                self.drawer.ctx.restore()
                self.drawer.draw(offsetx=self.offset, offsety=self.offsety, width=self.length)
        else:
            # Fallback to text
            self.text = self.current_layout[0].upper()
            base._TextBox.draw(self)

    def _get_layout_names(self):
        """
        Returns a sequence of tuples of layout name and lowercased class name
        strings for each available layout.
        """

        layouts = itertools.chain(
            self.qtile.config.layouts,
            (layout for group in self.qtile.config.groups for layout in group.layouts),
        )

        return set((layout.name, layout.__class__.__name__.lower()) for layout in layouts)

    def _update_icon_paths(self):
        self.icon_paths = []

        # We allow user to override icon search path
        self.icon_paths.extend(os.path.expanduser(path) for path in self.custom_icon_paths)

        # We also look in ~/.icons/ and ~/.local/share/icons
        self.icon_paths.append(os.path.expanduser("~/.icons"))
        self.icon_paths.append(os.path.expanduser("~/.local/share/icons"))

        # Default icons are in libqtile/resources/layout-icons.
        # If using default config without any custom icons,
        # this path will be used.
        root = os.sep.join(os.path.abspath(__file__).split(os.sep)[:-2])
        self.icon_paths.append(os.path.join(root, "resources", "layout-icons"))

    def find_icon_file_path(self, layout_name):
        for icon_path in self.icon_paths:
            for extension in ["png", "svg"]:
                icon_filename = "layout-{}.{}".format(layout_name, extension)
                icon_file_path = os.path.join(icon_path, icon_filename)
                if os.path.isfile(icon_file_path):
                    return icon_file_path

    def _setup_images(self):
        """
        Loads layout icons.
        """
        for names in self._get_layout_names():
            layout_name = names[0]
            # Python doesn't have an ordered set but we can use a dictionary instead
            # First key is the layout's name (which may have been set by the user),
            # the second is the class name. If these are the same (i.e. the user hasn't
            # set a name) then there is only one key in the dictionary.
            layouts = dict.fromkeys(names)
            for layout in layouts.keys():
                icon_file_path = self.find_icon_file_path(layout)
                if icon_file_path:
                    break
            else:
                logger.warning('No icon found for layout "%s"', layout_name)
                icon_file_path = self.find_icon_file_path("unknown")

            img = Img.from_path(icon_file_path)

            new_height = (self.bar.height - 2) * self.scale
            img.resize(height=new_height)
            if img.width > self.length:
                self.length = img.width + self.actual_padding * 2

            self.surfaces[layout_name] = img

        self.icons_loaded = True

    def finalize(self):
        self._remove_hooks()
        base._TextBox.finalize(self)



# Copyright (c) 2017 numirias
# Copyright (c) 2024 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import time
from collections import defaultdict

import pytest
from pytest import approx

from libqtile.config import Screen
from libqtile.confreader import Config
from libqtile.layout.plasma import AddMode, Node, NotRestorableError, Orient, Plasma, Priority
from test.layouts.layout_utils import assert_focused

Node.min_size_default = 10


class Canvas:
    horizontal_line = "\u2500"
    vertical_line = "\u2502"
    tl_corner = "\u250c"
    tr_corner = "\u2510"
    bl_corner = "\u2514"
    br_corner = "\u2518"

    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.canvas = defaultdict(lambda: defaultdict(lambda: "#"))

    def add_box(self, x, y, width, height, name="*"):
        width = width - 1
        height = height - 1
        label = str(name)[: width - 1]
        for i in range(x, width + x):
            self.canvas[i][y] = self.horizontal_line
            self.canvas[i][y + height] = self.horizontal_line
        for i in range(y, height + y):
            self.canvas[x][i] = self.vertical_line
            self.canvas[x + width][i] = self.vertical_line
        for i in range(x + 1, width + x):
            for j in range(y + 1, y + height):
                self.canvas[i][j] = "."
        for i, char in enumerate(label):
            self.canvas[x + 1 + i][y + 1] = char
        self.canvas[x][y] = self.tl_corner
        self.canvas[x + width][y] = self.tr_corner
        self.canvas[x][y + height] = self.bl_corner
        self.canvas[x + width][y + height] = self.br_corner

    def view(self):
        res = ""
        for y in range(self.height):
            for x in range(self.width):
                res += self.canvas[x][y]
            res += "\n"
        return res


def tree(node, level=0):
    res = "{indent}{name} {orient} {repr_} {pos} {size} {parent}\n".format(
        indent=level * 4 * " ",
        name="%s" % (node.payload or "*"),
        orient="H" if node.horizontal else "V",
        repr_="%s" % repr(node),
        pos="%g*%g@%g:%g" % (node.width, node.height, node.x, node.y),
        size="size: %s%s" % (node.size, " (auto)" if node.flexible else ""),
        parent="p: %s" % node.parent,
    )
    for child in node:
        res += tree(child, level + 1)
    return res


def draw(root):
    canvas = Canvas(root.width, root.height)

    def add(node):
        if node.is_leaf:
            canvas.add_box(*node.pixel_perfect, node.payload)
        for child in node:
            add(child)

    add(root)
    return canvas.view()


def info(node):
    print(tree(node))
    print(draw(node))


def create_nodes(string):
    for x in string.split():
        yield Node(x)


@pytest.fixture
def root():
    root = Node("root", 0, 0, 120, 50)
    return root


@pytest.fixture
def tiny_grid(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    return a, b, c


@pytest.fixture
def small_grid(root):
    a, b, c, d = create_nodes("a b c d")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    c.flip_with(d)
    return a, b, c, d


@pytest.fixture
def grid(root):
    a, b, c, d, e = create_nodes("a b c d e")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    c.flip_with(d)
    c.parent.add_child(e)
    return a, b, c, d, e


@pytest.fixture
def complex_grid(root):
    a, b, c, d, e, f, g = create_nodes("a b c d e f g")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    c.flip_with(d)
    c.parent.add_child(e)
    c.flip_with(f)
    f.flip_with(g)
    return a, b, c, d, e, f, g


@pytest.fixture(autouse=True)
def reset_node_priority():
    yield
    Node.priority = Priority.FIXED


##############
# TEST NODES #
##############


def test_single_node():
    n = Node(None, 0, 0, 120, 50)
    assert n.x == 0
    assert n.y == 0
    assert n.width == 120
    assert n.height == 50
    assert n.is_root is True
    assert n.is_leaf is True
    assert n.parent is None
    assert n.children == []
    assert n.orient == Orient.HORIZONTAL
    assert n.horizontal is True
    assert n.vertical is False
    assert n.size is None
    assert (n.x, n.y) == n.pos


def test_add_child(root):
    """Single child fills the entire root node's space."""
    child = Node("a")
    root.add_child(child)
    assert root.children == [child]
    assert child.parent == root
    assert root.width == child.width == 120
    assert root.height == child.height == 50
    assert root.x == child.x == 0
    assert root.y == child.y == 0


def test_add_children(root):
    """Multiple children will split the root node's space."""
    a, b = create_nodes("a b")
    root.add_child(a)
    root.add_child(b)
    assert root.width == 120
    assert a.width == b.width == 60
    assert root.height == 50
    assert a.height == b.height == 50
    assert a.pos == (0, 0)
    assert b.pos == (60, 0)
    c = Node("c")
    root.add_child(c)
    assert a.width == b.width == c.width == 40
    assert a.pos == (0, 0)
    assert b.pos == (40, 0)
    assert c.pos == (80, 0)


def test_add_child_after(root, grid):
    """Test adding nodes in specific positions."""
    a, b, c, d, e = grid
    f = Node("f")
    g = Node("g")
    h = Node("h")
    # add_child is called on the target node's parent
    # i.e. the node containing the target
    root.add_child_after(f, a)
    assert root.tree == [a, f, [b, [c, d, e]]]
    b.parent.add_child_after(g, b)
    assert root.tree == [a, f, [b, g, [c, d, e]]]
    d.parent.add_child_after(h, d)
    assert root.tree == [a, f, [b, g, [c, d, h, e]]]


def test_add_child_after_with_sizes(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    a.size += 10
    b.size += 10
    b.parent.add_child_after(c, b)
    assert a.size == b.size == 55
    assert c.size == 10


def test_remove_child(root):
    a, b = create_nodes("a b")
    root.add_child(a)
    root.add_child(b)
    root.remove_child(a)
    assert root.children == [b]
    root.remove_child(b)
    assert root.children == []


def test_nested(root):
    a, b, c, d = create_nodes("a b c d")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    assert a.width == 60
    assert b.width == 60
    assert c.width == 60
    assert a.height == 50
    assert b.height == 25
    assert c.height == 25
    b.flip_with(d)
    assert b.width == 30
    assert d.width == 30


def test_leaves(root, grid):
    a, b, c, d, e = grid
    assert root.first_leaf is a
    assert root.last_leaf is e
    assert b.parent.first_leaf is b
    assert b.parent.last_leaf is e


def test_directions(root, grid):
    a, b, c, d, e = grid
    assert a.up is None
    assert a.right is b
    assert a.down is None
    assert a.left is None

    assert b.up is None
    assert b.right is None
    assert b.down is c
    assert b.left is a

    assert c.up is b
    assert c.right is d
    assert c.down is None
    assert c.left is a

    assert d.up is b
    assert d.right is e
    assert d.down is None
    assert d.left is c

    assert e.up is b
    assert e.right is None
    assert e.down is None
    assert e.left is d


def test_prev_next(grid):
    a, b, c, d, e = grid
    assert a.next_leaf == b
    assert b.next_leaf == c
    assert c.next_leaf == d
    assert d.next_leaf == e
    assert e.next_leaf == a
    assert a.prev_leaf == e
    assert e.prev_leaf == d
    assert d.prev_leaf == c
    assert c.prev_leaf == b
    assert b.prev_leaf == a


def test_siblings(grid):
    a, b, c, d, e = grid
    assert d.siblings == [c, e]
    assert b.siblings == [c.parent]


def test_root_siblings(root):
    siblings = root.siblings
    assert siblings == []


def test_move_forward(root, grid):
    a, b, c, d, e = grid
    assert c.parent.children == [c, d, e]
    c.move_right()
    assert c.parent.children == [d, c, e]
    c.move_right()
    assert c.parent.children == [d, e, c]
    c.move_right()
    assert root.tree == [a, [b, [d, e]], c]


def test_move_backward(root, grid):
    a, b, c, d, e = grid
    e.move_left()
    assert c.parent.children == [c, e, d]
    e.move_left()
    assert c.parent.children == [e, c, d]
    e.move_left()
    assert root.tree == [a, e, [b, [c, d]]]


def test_advanced_move(grid):
    a, b, c, d, e = grid
    c.move_up()
    assert b.parent.tree == [b, c, [d, e]]
    a.move_up()
    assert b.parent.tree == [b, c, [d, e]]


def test_advanced_move2(root, grid):
    a, b, c, d, e = grid
    res = c.move_down()
    assert b.parent.tree == [b, [d, e], c]
    assert res is True
    res = e.move_down()
    assert b.parent.tree == [b, d, e, c]
    assert res is True
    res = e.move_left()
    assert root.tree == [a, e, [b, d, c]]
    assert res is True
    res = d.move_right()
    assert root.tree == [a, e, [b, c], d]
    assert res is True
    res = a.move_left()
    assert root.tree == [a, e, [b, c], d]
    assert res is False
    res = d.move_right()
    assert root.tree == [a, e, [b, c], d]
    assert res is False


def test_move_blocked(root, grid):
    a, b, c, d, e = grid
    orig_tree = root.tree.copy()
    res = a.move_up()
    assert root.tree == orig_tree
    assert res is False
    res = b.move_up()
    assert root.tree == orig_tree
    assert res is False


def test_move_root(root):
    a = Node("a")
    root.add_child(a)
    root.move_up()
    assert root.tree == [a]


def test_integrate(root):
    a, b, c, d, e = create_nodes("a b c d e")
    root.add_child(a)
    root.add_child(b)
    root.add_child(c)
    root.add_child(d)
    c.integrate_left()
    assert root.tree == [a, [b, c], d]
    a.integrate_right()
    assert root.tree == [[b, c, a], d]
    a.parent.add_child(e)
    c.integrate_down()
    assert root.tree == [[b, [a, c], e], d]
    e.integrate_up()
    assert root.tree == [[b, [a, c, e]], d]


def test_integrate_nested(root, grid):
    a, b, c, d, e = grid
    c.integrate_right()
    assert root.tree == [a, [b, [[d, c], e]]]


def test_move_and_integrate(root, grid):
    a, b, c, d, e = grid
    c.integrate_left()
    assert root.tree == [[a, c], [b, [d, e]]]
    a.integrate_right()
    assert root.tree == [c, [b, [d, e], a]]
    d.integrate_down()
    assert root.tree == [c, [b, e, [a, d]]]
    a.integrate_up()
    assert root.tree == [c, [b, [e, a], d]]
    e.integrate_left()
    assert root.tree == [[c, e], [b, a, d]]
    f = Node("f")
    a.flip_with(f)
    g = Node("g")
    a.flip_with(g)
    g.integrate_left()
    assert root.tree == [[c, e, g], [b, [a, f], d]]


def test_impossible_integrate(root, grid):
    a, b, c, d, e = grid
    orig_tree = root.tree.copy()
    a.integrate_left()
    assert orig_tree == root.tree
    b.integrate_up()
    assert orig_tree == root.tree


def test_impossible_integrate2(root):
    a, b = create_nodes("a b")
    root.add_child(a)
    root.add_child(b)
    orig_tree = root.tree.copy()
    b.integrate_up()
    assert root.tree == orig_tree
    b.integrate_down()
    assert root.tree == orig_tree
    b.integrate_right()
    assert root.tree == orig_tree
    a.integrate_up()
    assert root.tree == orig_tree
    a.integrate_down()
    assert root.tree == orig_tree
    a.integrate_left()
    assert root.tree == orig_tree


def test_find_payload(root, grid):
    a, b, c, d, e = grid
    assert root.find_payload(a.payload) is a
    assert root.find_payload(b.payload) is b
    assert root.find_payload(d.payload) is d
    assert root.find_payload("x") is None


def test_last_access(grid):
    a, b, c, d, e = grid
    f = Node("f")
    a.flip_with(f)
    d.access()
    assert b.down is d
    b.access()
    assert f.right is b
    f.access()
    assert b.left is f


def test_root_without_dimensions():
    """A root node with undef. dimensions should be able to add a child."""
    root = Node()
    x = Node("x")
    root.add_child(x)


def test_root(root, grid):
    for node in grid:
        assert node.root is root


def test_all(root, grid):
    assert set(root.all_leafs) == set(grid)


def test_close_neighbor(root):
    a, b, c, d = create_nodes("a b c d")
    root.add_child(a)
    root.add_child(b)
    a.flip_with(c)
    b.flip_with(d)
    assert a.close_up is None
    assert a.close_left is None
    assert a.close_right is b
    assert a.close_down is c

    assert b.close_up is None
    assert b.close_left is a
    assert b.close_right is None
    assert b.close_down is d

    assert c.close_up is a
    assert c.close_left is None
    assert c.close_right is d
    assert c.close_down is None

    assert d.close_up is b
    assert d.close_left is c
    assert d.close_right is None
    assert d.close_down is None


def test_close_neighbor2(root, small_grid):
    a, b, c, d = small_grid
    assert b.close_left is a


def test_close_neighbor_nested(root, grid):
    a, b, c, d, e = grid
    f, g, h, i, j, k, m = create_nodes("f g h i j k m")
    root.add_child(f)
    d.flip_with(h)
    a.flip_with(i)
    e.flip_with(j)
    e.parent.add_child(k)
    f.flip_with(m)
    f.height = 10
    assert b.close_down is d
    b.flip_with(g)
    assert b.close_down is c
    assert d.close_right is e
    assert e.close_left is d
    assert m.close_left is e
    assert e.close_up is g
    assert m.close_right is None
    assert h.close_down is None


def test_close_neighbor_approx(root, small_grid):
    """Tolerate floating point errors when calculating common borders."""
    root.height += 30
    a, b, c, d = small_grid
    e, f, g = create_nodes("e f g")
    c.flip_with(f)
    b.parent.add_child(e)
    c.parent.add_child(g)
    assert g.close_down is e


def test_points(root, small_grid):
    a, b, c, d = small_grid
    assert c.x, c.y == (60, 25)
    assert c.x + c.width, c.y == (90, 25)
    assert c.x, c.y + c.height == (60, 50)
    assert c.x + c.width, c.y + c.height == (90, 50)


def test_center(root):
    assert root.x_center == 60
    assert root.y_center == 25
    assert root.center == (60, 25)


def test_recent_leaf(root, grid):
    a, b, c, d, e = grid
    assert d.parent.recent_leaf is c
    c.access()
    d.access()
    assert d.parent.recent_leaf is d
    b.access()
    c.access()
    assert root.recent_leaf is c
    a.access()
    assert root.recent_leaf is a


def test_recent_close_neighbor(root, grid):
    a, b, c, d, e = grid
    assert b.close_down is d
    c.access()
    assert b.close_down is c
    assert a.close_right is c
    b.access()
    assert a.close_right is b


def test_add_node(root):
    a, b, c, d, e, f, g = create_nodes("a b c d e f g")
    root.add_node(a)
    assert root.tree == [a]
    root.add_node(b)
    assert root.tree == [a, b]
    a.add_node(c)
    assert root.tree == [a, c, b]
    c.add_node(d, mode=AddMode.HORIZONTAL)
    assert root.tree == [a, c, d, b]
    root.remove_child(d)
    c.add_node(d, mode=AddMode.VERTICAL)
    c.parent.add_child_after
    assert root.tree == [a, [c, d], b]
    c.add_node(e, mode=AddMode.VERTICAL)
    assert root.tree == [a, [c, e, d], b]
    assert a.width == 40
    a.add_node(f, mode=AddMode.HORIZONTAL | AddMode.SPLIT)
    assert root.tree == [a, f, [c, e, d], b]
    assert a.width == f.width == 20
    assert c.parent.width == b.width == 40
    a.add_node(g, mode=AddMode.VERTICAL | AddMode.SPLIT)
    assert root.tree == [[a, g], f, [c, e, d], b]


def test_contains(root, grid):
    x = Node("x")
    nodes = list(grid)
    nodes += [n.parent for n in nodes]
    nodes.append(root)
    for n in nodes:
        assert n in root
    assert x not in root


def test_size(grid):
    a, b, c, d, e = grid
    assert a.size == a.width == 60
    assert b.size == b.height == 25


def test_capacity(root, grid):
    a, b, c, d, e = grid
    assert root.capacity == 120
    assert b.parent.capacity == 50
    assert c.parent.capacity == 60
    assert c.capacity == 25


def test_capacity2(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)


def test_resize(root, grid):
    a, b, c, d, e = grid
    a.size += 10
    assert a.width == a.size == 70
    assert b.height == b.size == 25
    assert b.width == 50
    assert c.width == d.width == e.width == 50 / 3
    assert a.pos == (0, 0)
    assert b.pos == (70, 0)
    assert c.pos == (70, 25)
    assert d.pos == (70 + 50 / 3, 25)
    assert e.pos == (70 + (50 / 3) * 2, 25)
    b.size -= 5
    assert c.width == d.width == e.width == 50 / 3
    assert c.height == d.height == e.height == 30
    d.size += 5
    assert d.width == 50 / 3 + 5
    d.move_up()
    assert d.size == (50 - b.size) / 2
    b.integrate_down()
    assert b.size == d.size == 25
    assert b.parent.size == 25


def test_resize_absolute(grid):
    a, b, c, d, e = grid
    b.size = 10
    assert b.size == b.height == 10
    assert c.parent.size == 40
    b.size = 5
    assert b.size == 10


def test_resize_absolute2(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    root.add_child(c)
    a.size = 30
    b.size = 60
    c.size = 40
    assert a.size == 30 * (80 / 90)
    assert b.size == 60 * (80 / 90)
    assert c.size == 40


def test_resize_absolute_and_relative(root):
    a, b, c, d = create_nodes("a b c d")
    root.add_child(a)
    root.add_child(b)
    a.size = 20
    b.size = 20
    assert a.size == 100
    assert b.size == 20
    root.add_child(c)
    assert c.size == approx(10)
    assert a.size == approx(100 * (11 / 12))
    assert b.size == approx(20 * (11 / 12))
    root.add_child(d)
    assert c.size == d.size == approx(10)


def test_resize_absolute_and_relative_balanced(root):
    Node.priority = Priority.BALANCED
    a, b, c, d = create_nodes("a b c d")
    root.add_child(a)
    root.add_child(b)
    a.size = 20
    b.size = 20
    assert a.size == 100
    assert b.size == 20
    root.add_child(c)
    assert c.size == approx(40)
    assert a.size == approx(100 * (2 / 3))
    assert b.size == approx(20 * (2 / 3))
    root.add_child(d)
    assert c.size == d.size == approx(20)


def test_resize_absolute_and_relative2(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    root.add_child(c)
    a.size += 10
    assert a.size == 50
    assert b.size == 35
    assert c.size == 35
    b.size += 10
    assert a.size == 50
    assert b.size == 45
    assert c.size == 25


def test_resize_flat(root):
    a, b, c, d, e, f = create_nodes("a b_abs c d e_abs f_abs")
    root.add_child(a)
    root.add_child(b)
    root.add_child(c)
    root.add_child(d)
    d.flip_with(e)
    e.flip_with(f)
    b.size = b.size
    e.size = e.size
    f.size = f.size
    a.size = 60
    assert a.size == 60
    assert b.size == 25
    assert c.size == 10
    assert d.parent.size == 25
    assert e.size == f.size == 25 / 2


def test_resize_minimum(grid):
    a, b, c, d, e = grid
    b.size -= 100
    assert b.size == 10


def test_resize_all_absolute_underflow(root, grid):
    a, b, c, d, e = grid
    c.size = 10
    d.size = 10
    assert e.size == 40
    e.size = 10
    assert e.size == 10
    assert c.size == d.size == 25


def test_resize_all_absolute_overflow(grid):
    a, b, c, d, e = grid
    c.size = d.size = 15
    e.size = 40
    assert e.size == 40
    assert c.size == d.size == 10
    e.size = 50
    assert e.size == 40
    assert c.size == d.size == 10


def test_resize_overflow_with_relative(root, grid):
    a, b, c, d, e = grid
    c.size = 20
    d.size = 40
    assert c.size == 10
    assert d.size == 40
    assert e.size == 10
    assert e.flexible
    d.size = 50
    assert c.size == 10
    assert d.size == 40
    assert e.size == 10
    assert e.flexible


def test_resize_overflow_with_relative2(root, grid):
    a, b, c, d, e = grid
    c.size = 20
    d.size = 20
    a.size = 70
    assert a.size == 70
    assert c.size == d.size == 20
    assert e.size == 10
    a.size = 80
    assert a.size == 80
    assert c.size == d.size == 15
    assert e.size == 10
    a.size = 90
    assert a.size == 90
    assert c.size == d.size == e.size == 10
    a.size = 100
    assert a.size == 90


def test_resize_only_absolute_remains(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    a.size = 20
    b.size = 20
    root.add_child(c)
    root.remove_child(c)
    assert a.size == 100
    assert b.size == 20


def test_reset_size(grid):
    a, b, c, d, e = grid
    a.size += 5
    assert a.size == 65
    a.reset_size()
    assert a.size == 60


def test_size_after_split(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    b.size -= 20
    b.flip_with(c)
    assert b.parent.size == 40
    assert b.size == c.size == 25
    b.remove()
    assert c.size == 40


def test_only_child_must_be_flexible(root):
    a, b = create_nodes("a b")
    root.add_child(a)
    root.add_child(b)
    a.size = 10
    root.remove_child(b)
    assert a.flexible


def test_deny_only_child_resize(root):
    a = Node("a")
    root.add_child(a)
    a.size = 10
    assert a.size == 120


def test_resize_parents(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    b.width += 10
    assert b.parent.size == 70
    assert b.size == c.size == 25


def test_pixelperfect(root, tiny_grid):
    a, b, c = tiny_grid
    root._height = 11
    root._width = 11
    ds = a.pixel_perfect
    assert all(type(x) is int for x in (ds.x, ds.y, ds.width, ds.height))
    assert a.width + b.width == 11
    assert a.pixel_perfect.width + b.pixel_perfect.width == 11
    assert b.height + c.height == 11
    assert b.pixel_perfect.height + c.pixel_perfect.height == 11


def test_pixelperfect_draw(root, complex_grid):
    root._height = 10
    for i in range(40, 50):
        root._width = i
        view = draw(root)
        assert "#" not in view
    root._width = 50
    for i in range(10, 20):
        root._height = i
        view = draw(root)
        assert "#" not in view


def test_resize_root(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    a.height += 10
    root.height += 10
    root.width += 10
    root.size = 10
    assert a._size is b._size is root._size is None


def test_set_xy(root, tiny_grid):
    a, b, c = tiny_grid
    root.x = 10
    root.y = 20
    assert root.x == 10
    assert root.y == 20
    a.x = 30
    a.y = 40
    assert a.x == root.x == 10
    assert a.y == root.y == 20
    root.width = 50
    root.height = 60
    assert root._width == 50
    assert root._height == 60


def test_set_width_height(root, tiny_grid):
    a, b, c = tiny_grid
    a.width = 70
    assert a.width == 70
    assert b.width == c.width == 50
    b.height = 30
    assert b.height == 30
    assert c.height == 20
    b.width = 80
    assert b.width == c.width == 80
    assert a.width == 40
    a.height = 20
    assert a.height == 50


def test_min_size(root, small_grid):
    a, b, c, d = small_grid
    c.size += 10
    d.size += 20
    b.size = 20
    assert a.min_size == Node.min_size_default
    assert b.parent.min_size == 60
    assert b.min_size == 20
    assert c.parent.min_size == Node.min_size_default
    assert c.min_size == 20
    assert d.min_size == 40


def test_transitive_flexible(root, complex_grid):
    a, b, c, d, e, f, g = complex_grid
    assert b.parent.flexible
    d.size = 20
    e.size = 20
    f.size = 10
    assert b.parent.flexible
    g.size = 10
    assert not b.parent.flexible


def test_resize_bubbles(root, small_grid):
    a, b, c, d = small_grid
    c.size += 10
    d.size += 20
    assert c.size == 20
    assert d.size == 40
    a.size = 30
    assert c.size == 30
    assert d.size == 60


def test_resize_bubbles2(root, complex_grid):
    a, b, c, d, e, f, g = complex_grid
    c.flip_with(Node("h"))
    f.size += 10
    g.size += 10
    assert f.size == g.size == 10
    assert f.fixed and g.fixed
    assert d.size == e.size == 20
    assert d.flexible and e.flexible
    a.size -= 40
    assert a.size == 20
    assert f.size == g.size == 10
    assert d.size == e.size == 40
    d.size = 10
    assert d.size == 10
    assert e.size == 70
    assert f.size == g.size == 10
    assert e.flexible
    e.size = 10
    assert e.fixed


def test_resize_bubbles3(root, complex_grid):
    a, b, c, d, e, f, g = complex_grid
    h = Node("h")
    c.flip_with(h)
    f.size += 10
    g.size += 10
    assert f.size == g.size == c.size == h.size == 10
    a.size = 10
    assert a.size == 10
    assert f.size == g.size == c.size == h.size == 10
    assert d.size == e.size == 45
    d.size = 10
    assert d.size == 10
    assert e.size == 80
    e.size = 10
    assert e.size == 10
    assert f.size == g.size == c.size == h.size == d.size == 100 / 3


def test_resize_nested(root):
    a, b, c, d, e, f, g, h = create_nodes("a b c_abs d_abs e f g h_abs")
    nu1, nu2, nd, mu, md1, md2 = create_nodes("nu1_abs nu2_abs nd mu md1_abs md2")
    ou1, ou2, od, pu, pd1, pd2 = create_nodes("ou1_abs ou2_abs od pu pd1_abs " "pd2_abs")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    b.parent.add_child(e)
    b.parent.add_child(g)
    c.flip_with(d)
    e.flip_with(f)
    g.flip_with(h)

    b.parent.add_child(nu1)
    nu1.flip_with(mu)
    nu1.flip_with(nd)
    nu1.flip_with(nu2)
    mu.flip_with(md1)
    md1.flip_with(md2)

    b.parent.add_child(ou1)
    ou1.flip_with(pu)
    ou1.flip_with(od)
    ou1.flip_with(ou2)
    pu.flip_with(pd1)
    pd1.flip_with(pd2)

    def assert_first_state():
        assert b.parent.size == 60
        assert c.size == approx(40)
        assert d.size == 20
        assert e.size == f.size == 30
        assert g.size == 40
        assert h.size == 20
        assert nu1.size == 10
        assert nu2.size == 20
        assert nd.parent.size == 30
        assert mu.parent.size == 30
        assert md1.size == 10
        assert md2.size == 20
        assert ou1.size == 10
        assert ou2.size == 20
        assert od.parent.size == 30
        assert pu.parent.size == 30
        assert pd1.size == 10
        assert pd2.size == 20

    def assert_second_state():
        assert a.size == 30
        assert b.parent.size == 90
        assert c.size == 60
        assert d.size == 30
        assert e.size == f.size == 45
        assert g.size == 70
        assert h.size == 20
        assert nu1.size == 10
        assert nu2.size == 20
        assert nd.parent.size == 30
        assert mu.parent.size == 60
        assert md1.size == 10
        assert md2.size == 50
        assert ou1.size == 15
        assert ou2.size == 30
        assert od.parent.size == 45
        assert pd1.size == 15
        assert pd2.size == 30
        assert pu.parent.size == 45

    b.parent.size = 60
    c.size += 5
    d.size -= 5
    h.size = 20
    nu1.size = 10
    nu2.size = 20
    md1.size = 10
    ou1.size = 10
    ou2.size = 20
    pd1.size = 10
    pd2.size = 20

    assert a.size == 60
    assert_first_state()
    a.size -= 30
    assert_second_state()
    a.size += 30
    assert a.size == 60
    assert_first_state()
    b.parent.size += 30
    assert_second_state()
    b.parent.size -= 30
    assert a.size == 60
    assert_first_state()


def test_resize_nested_balanced(root):
    Node.priority = Priority.BALANCED
    a, b, c, d, e, f, g, h = create_nodes("a b c_abs d_abs e f g h_abs")
    nu1, nu2, nd, mu, md1, md2 = create_nodes("nu1_abs nu2_abs nd mu md1_abs md2")
    ou1, ou2, od, pu, pd1, pd2 = create_nodes("ou1_abs ou2_abs od pu pd1_abs " "pd2_abs")
    root.add_child(a)
    root.add_child(b)
    b.flip_with(c)
    b.parent.add_child(e)
    b.parent.add_child(g)
    c.flip_with(d)
    e.flip_with(f)
    g.flip_with(h)

    b.parent.add_child(nu1)
    nu1.flip_with(mu)
    nu1.flip_with(nd)
    nu1.flip_with(nu2)
    mu.flip_with(md1)
    md1.flip_with(md2)

    b.parent.add_child(ou1)
    ou1.flip_with(pu)
    ou1.flip_with(od)
    ou1.flip_with(ou2)
    pu.flip_with(pd1)
    pd1.flip_with(pd2)

    def assert_first_state():
        assert b.parent.size == 60
        assert c.size == 40
        assert d.size == 20
        assert e.size == f.size == 30
        assert g.size == 40
        assert h.size == 20
        assert nu1.size == 10
        assert nu2.size == 20
        assert nd.parent.size == 30
        assert mu.parent.size == 30
        assert md1.size == 10
        assert md2.size == 20
        assert ou1.size == 10
        assert ou2.size == 20
        assert od.parent.size == 30
        assert pu.parent.size == 30
        assert pd1.size == 10
        assert pd2.size == 20

    def assert_second_state():
        assert a.size == 30
        assert b.parent.size == 90
        assert c.size == 60
        assert d.size == 30
        assert e.size == f.size == 45
        assert g.size == 70
        assert h.size == 20
        assert nu1.size == 10
        assert nu2.size == 20
        assert nd.parent.size == 30
        assert mu.parent.size == 60
        assert md1.size == 10
        assert md2.size == 50
        assert ou1.size == 15
        assert ou2.size == 30
        assert od.parent.size == 45
        assert pd1.size == 15
        assert pd2.size == 30
        assert pu.parent.size == 45

    b.parent.size = 60
    c.size += 5
    d.size -= 5
    h.size = 20
    nu1.size = 10
    nu2.size = 20
    md1.size = 10
    ou1.size = 10
    ou2.size = 20
    pd1.size = 10
    pd2.size = 20

    assert a.size == 60
    assert_first_state()
    a.size -= 30
    assert_second_state()
    a.size += 30
    assert a.size == 60
    assert_first_state()
    b.parent.size += 30
    assert_second_state()
    b.parent.size -= 30
    assert a.size == 60
    assert_first_state()

    a.size = 30
    x = Node("x")
    root.add_child(x)
    assert x.size == 40
    assert_first_state()
    x.remove()
    assert_second_state()

    a.remove()
    assert b.width == 120
    y = Node("y")
    root.add_child(y)
    assert_first_state()


def test_resize_max(root, tiny_grid):
    a, b, c = tiny_grid
    a.width = 120
    assert a.width == 110
    assert b.width == c.width == 10


def test_restore(root, grid):
    a, b, c, d, e = grid
    tree = root.tree
    for node in grid:
        node.remove()
        root.restore(node)
        assert root.tree == tree


def test_restore_same_payload(root, grid):
    """Restore a node that's not identical with the removed one but carries
    the same payload.
    """
    a, b, c, d, e = grid
    d.remove()
    new_d = Node("d")
    root.restore(new_d)
    assert root.tree == [a, [b, [c, new_d, e]]]


def test_restore_unknown(root, grid):
    a, b, c, d, e = grid
    with pytest.raises(NotRestorableError):
        root.restore(Node("x"))
    d.remove()
    with pytest.raises(NotRestorableError):
        root.restore(Node("x"))
    root.restore(d)
    assert root.tree == [a, [b, [c, d, e]]]


def test_restore_no_parent(root, small_grid):
    a, b, c, d = small_grid
    c.remove()
    d.remove()
    with pytest.raises(NotRestorableError):
        root.restore(c)
    root.restore(d)
    assert root.tree == [a, [b, d]]


def test_restore_bad_index(root, grid):
    a, b, c, d, e = grid
    f, g = create_nodes("f g")
    e.parent.add_child(f)
    e.parent.add_child(g)
    g.remove()
    f.remove()
    e.remove()
    root.restore(g)
    assert root.tree == [a, [b, [c, d, g]]]


def test_restore_sizes(root, grid):
    a, b, c, d, e = grid
    c.size = 30
    c.remove()
    root.restore(c)
    assert c.size == 30
    c.remove()
    d.size = 30
    e.size = 30
    assert d.size == e.size == 30
    root.restore(c)
    assert c.size == 30
    assert d.size == e.size == 15


def test_restore_sizes_flip(root, tiny_grid):
    a, b, c = tiny_grid
    c.size = 10
    c.remove()
    assert a._size is b._size is None
    root.restore(c)
    assert c.size == 10
    b.size = 10
    c.remove()
    root.restore(c)
    assert b.size == 10
    b.remove()
    root.restore(b)
    assert b.size == 10


def test_restore_root(root):
    a, b = create_nodes("a b")
    root.add_child(a)
    root.add_child(b)
    a.size = 20
    a.remove()
    root.restore(a)
    assert a._size == 20
    assert b._size is None
    b.remove()
    root.restore(b)
    assert a._size == 20
    assert b._size is None


def test_restore_root2(root):
    a, b, c = create_nodes("a b c")
    root.add_child(a)
    root.add_child(b)
    root.add_child(c)
    b.size = 20
    c.size = 40
    a.remove()
    assert b.size == 40
    assert c.size == 80
    root.restore(a)
    assert not a.fixed
    assert a.size == approx(60)
    assert b.size == approx(20)
    assert c.size == approx(40)


def test_restore_keep_flexible(root, tiny_grid):
    a, b, c = tiny_grid
    b.remove()
    root.restore(b)
    assert a._size is b._size is c._size is None
    b.size = 10
    b.remove()
    root.restore(b)
    assert b._size == 10
    assert c._size is None
    c.remove()
    root.restore(c)
    assert b._size == 10
    assert c._size is None
    c.size = 10
    b.reset_size()
    b.remove()
    root.restore(b)
    assert b._size is None
    assert c._size == 10
    c.remove()
    root.restore(c)
    assert b._size is None
    assert c._size == 10


def test_resize_with_collapse_and_restore(root, small_grid):
    a, b, c, d = small_grid
    root.height = 30
    c.size = 30
    d.size += 10
    b.remove()
    assert c.size == c.height == 10
    assert d.size == d.height == 20
    root.restore(b)
    assert b.height == 15
    assert b.width == 60
    assert c.height == d.height == 15
    assert c.width == 20
    assert d.width == 40


def test_node_repr(root, grid):
    a, b, c, d, e = grid
    assert repr(root).startswith("<Node root +2 ")
    assert repr(a).startswith("<Node a ")
    assert repr(b).startswith("<Node b ")
    assert repr(c).startswith("<Node c ")
    assert repr(d).startswith("<Node d ")
    assert repr(e).startswith("<Node e ")


################
# LAYOUT TESTS #
################


@pytest.fixture
def load_grid(request):
    return getattr(request, "param", False)


@pytest.fixture
def plasma(manager_nospawn, request, load_grid):
    class PlasmaConfig(Config):
        layouts = [Plasma(**getattr(request, "param", dict()))]
        screens = [Screen()]

    manager_nospawn.start(PlasmaConfig)

    if load_grid:
        manager_nospawn.test_window("a")
        manager_nospawn.test_window("b")
        manager_nospawn.c.layout.previous()
        manager_nospawn.c.layout.mode_vertical()
        manager_nospawn.test_window("c")
        manager_nospawn.c.layout.right()
        manager_nospawn.c.layout.mode_vertical()
        manager_nospawn.test_window("d")

    yield manager_nospawn


def config(**kwargs):
    return pytest.mark.parametrize("plasma", [kwargs], indirect=True)


with_grid = pytest.mark.parametrize("load_grid", [True], indirect=True)


class Client:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


def client_tree(plasma):
    return plasma.c.layout.info()["tree"]


def test_init():
    layout = Plasma()
    assert isinstance(layout.root, Node)


def test_focus(root):
    layout = Plasma()
    layout.root = root
    a, b, c, d = [Client(x) for x in "abcd"]
    layout.add_client(a)
    layout.add_client(b)
    layout.add_client(c)
    layout.add_client(d)
    assert layout.focus_first() is a
    assert layout.focus_last() is d
    assert layout.focus_next(b) is c
    assert layout.focus_previous(c) is b
    layout.focus(c)
    assert layout.focused is c


def test_access(root):
    layout = Plasma()
    layout.root = root
    layout.add_client("a")
    now = time.time()
    assert layout.root.find_payload("a").last_accessed < now
    layout.focus("a")
    assert layout.root.find_payload("a").last_accessed > now


def test_info(plasma):
    plasma.test_window("a")
    plasma.test_window("b")
    assert client_tree(plasma) == ["a", "b"]


def test_windows(plasma):
    plasma.test_window("a")
    plasma.test_window("b")
    plasma.test_window("c")
    assert_focused(plasma, "c")
    assert client_tree(plasma) == ["a", "b", "c"]


def test_split_directions(plasma):
    plasma.test_window("a")
    plasma.c.layout.mode_horizontal()
    plasma.test_window("b")
    plasma.c.layout.mode_vertical()
    plasma.test_window("c")
    assert client_tree(plasma) == ["a", ["b", "c"]]


@with_grid
def test_layout_directions(plasma):
    assert_focused(plasma, "d")
    plasma.c.layout.left()
    assert_focused(plasma, "c")
    plasma.c.layout.up()
    assert_focused(plasma, "a")
    plasma.c.layout.right()
    assert_focused(plasma, "b")
    plasma.c.layout.down()
    assert_focused(plasma, "d")
    plasma.c.layout.down()
    assert_focused(plasma, "d")
    plasma.c.layout.previous()
    assert_focused(plasma, "b")
    plasma.c.layout.next()
    assert_focused(plasma, "d")


@with_grid
def test_move(plasma):
    assert client_tree(plasma) == [["a", "c"], ["b", "d"]]
    plasma.c.layout.move_up()
    assert client_tree(plasma) == [["a", "c"], ["d", "b"]]
    plasma.c.layout.move_down()
    assert client_tree(plasma) == [["a", "c"], ["b", "d"]]
    plasma.c.layout.move_left()
    assert client_tree(plasma) == [["a", "c"], "d", "b"]
    plasma.c.layout.move_right()
    assert client_tree(plasma) == [["a", "c"], "b", "d"]


@with_grid
def test_client_integrate(plasma):
    plasma.c.layout.integrate_left()
    assert client_tree(plasma) == [["a", "c", "d"], "b"]
    plasma.c.layout.integrate_up()
    assert client_tree(plasma) == [["a", ["c", "d"]], "b"]
    plasma.c.layout.integrate_up()
    plasma.c.layout.integrate_down()
    assert client_tree(plasma) == [["a", ["c", "d"]], "b"]
    plasma.c.layout.integrate_right()
    assert client_tree(plasma) == [["a", "c"], ["b", "d"]]


def test_sizes(plasma):
    Node.priority = Priority.BALANCED
    plasma.test_window("a")
    plasma.test_window("b")
    plasma.c.layout.mode_vertical()
    plasma.test_window("c")
    info = plasma.c.window.info()
    assert info["x"] == 400
    assert info["y"] == 300
    assert info["width"] == 400 - 2
    assert info["height"] == 300 - 2
    plasma.c.layout.grow_height(50)
    info = plasma.c.window.info()
    assert info["height"] == 300 - 2 + 50
    plasma.c.layout.grow_width(50)
    info = plasma.c.window.info()
    assert info["width"] == 400 - 2 + 50
    plasma.c.layout.reset_size()
    info = plasma.c.window.info()
    assert info["height"] == 300 - 2
    plasma.c.layout.set_height(300)
    info = plasma.c.window.info()
    assert info["height"] == 300 - 2
    plasma.c.layout.set_width(250)
    info = plasma.c.window.info()
    assert info["width"] == 250 - 2
    plasma.c.layout.set_size(200)
    info = plasma.c.window.info()
    assert info["height"] == 200 - 2
    plasma.c.layout.grow(10)
    info = plasma.c.window.info()
    assert info["height"] == 210 - 2


def test_remove(plasma):
    a = plasma.test_window("a")
    b = plasma.test_window("b")
    assert client_tree(plasma) == ["a", "b"]
    plasma.kill_window(a)
    assert client_tree(plasma) == ["b"]
    plasma.kill_window(b)
    assert client_tree(plasma) == []


def test_split_mode(plasma):
    plasma.test_window("a")
    plasma.test_window("b")
    plasma.c.layout.mode_horizontal_split()
    plasma.test_window("c")
    assert plasma.c.window.info()["width"] == 200 - 2
    plasma.c.layout.mode_vertical()
    plasma.test_window("d")
    assert plasma.c.window.info()["height"] == 300 - 2
    plasma.test_window("e")
    assert plasma.c.window.info()["height"] == 200 - 2
    plasma.c.layout.mode_vertical_split()
    plasma.test_window("f")
    assert plasma.c.window.info()["height"] == 100 - 2


def test_recent(plasma):
    plasma.test_window("a")
    plasma.test_window("b")
    plasma.test_window("c")
    assert_focused(plasma, "c")
    plasma.c.layout.recent()
    assert_focused(plasma, "b")
    plasma.c.layout.recent()
    assert_focused(plasma, "c")
    plasma.c.layout.next()
    assert_focused(plasma, "a")
    plasma.c.layout.recent()
    assert_focused(plasma, "c")


def test_bug_10():
    """Adding nodes when the correct root dimensions are still unknown
    should not raise an error.
    """
    layout = Plasma()
    layout.add_client(object())
    layout.add_client(object())



# Copyright (c) 2020 Guangwang Huang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
import subprocess

import pytest

import libqtile.bar
import libqtile.config
import libqtile.layout
import libqtile.widget
from libqtile.confreader import Config
from libqtile.lazy import lazy


class ServerConfig(Config):
    auto_fullscreen = True
    keys = [
        libqtile.config.Key(["mod4"], "Return", lazy.spawn("xterm")),
        libqtile.config.Key(["mod4"], "t", lazy.spawn("xterm"), desc="dummy description"),
        libqtile.config.Key([], "y", desc="noop"),
        libqtile.config.KeyChord(
            ["mod4"],
            "q",
            [
                libqtile.config.KeyChord(
                    [],
                    "q",
                    [
                        libqtile.config.Key([], "a", lazy.togroup("a")),
                    ],
                ),  # unnamed
                libqtile.config.Key([], "b", lazy.togroup("b")),
            ],
            mode="named",
        ),
    ]
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
    ]
    layouts = [
        libqtile.layout.Stack(num_stacks=1),
        libqtile.layout.Stack(num_stacks=2),
        libqtile.layout.Stack(num_stacks=3),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.TextBox(name="one"),
                ],
                20,
            ),
        ),
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.TextBox(name="two"),
                ],
                20,
            ),
        ),
    ]


server_config = pytest.mark.parametrize("manager", [ServerConfig], indirect=True)


def run_qtile_cmd(args, no_eval=False):
    cmd = os.path.join(os.path.dirname(__file__), "..", "bin", "qtile")
    argv = [cmd, "cmd-obj"]
    argv.extend(args.split())
    pipe = subprocess.Popen(argv, stdout=subprocess.PIPE)
    output, _ = pipe.communicate()
    output = output.decode()
    if no_eval:
        return output
    return eval(output)  # as returned by pprint.pprint


@server_config
def test_qtile_cmd(manager):
    manager.test_window("foo")
    wid = manager.c.window.info()["id"]

    for obj in ["window", "group", "screen"]:
        assert run_qtile_cmd("-s {} -o {} -f info".format(manager.sockfile, obj))

    layout = run_qtile_cmd("-s {} -o layout -f info".format(manager.sockfile))
    assert layout["name"] == "stack"
    assert layout["group"] == "a"

    window = run_qtile_cmd("-s {} -o window {} -f info".format(manager.sockfile, wid))
    assert window["id"] == wid
    assert window["name"] == "foo"
    assert window["group"] == "a"

    group = run_qtile_cmd("-s {} -o group {} -f info".format(manager.sockfile, "a"))
    assert group["name"] == "a"
    assert group["screen"] == 0
    assert group["layouts"] == ["stack", "stack", "stack"]
    assert group["focus"] == "foo"

    assert run_qtile_cmd("-s {} -o screen {} -f info".format(manager.sockfile, 0)) == {
        "height": 600,
        "index": 0,
        "width": 800,
        "x": 0,
        "y": 0,
    }

    bar = run_qtile_cmd("-s {} -o bar {} -f info".format(manager.sockfile, "bottom"))
    assert bar["height"] == 20
    assert bar["width"] == 800
    assert bar["size"] == 20
    assert bar["position"] == "bottom"


@server_config
def test_display_kb(manager):
    from pprint import pprint

    cmd = "-s {} -o root -f display_kb".format(manager.sockfile)
    table = run_qtile_cmd(cmd)
    print(table)
    pprint(table)
    assert table.count("\n") >= 2
    assert re.match(r"(?m)^Mode\s{3,}KeySym\s{3,}Mod\s{3,}Command\s{3,}Desc\s*$", table)
    assert re.search(r"(?m)^<root>\s{3,}Return\s{3,}mod4\s{3,}spawn\('xterm'\)\s*$", table)
    assert re.search(
        r"(?m)^<root>\s{3,}t\s{3,}mod4\s{3,}spawn\('xterm'\)\s{3,}dummy description\s*$", table
    )
    assert re.search(r"(?m)^<root>\s{3,}q\s{3,}mod4\s{13,}Enter named mode\s*$", table)
    assert re.search(r"(?m)^named\s{3,}q\s{13,}Enter <unnamed> mode\s*$", table)
    assert re.search(r"(?m)^named\s{3,}b\s{9,}togroup\('b'\)\s*$", table)
    assert re.search(r"(?m)^named>_\s{3,}a\s{9,}togroup\('a'\)\s*$", table)
    assert re.search(r"(?m)^<root>\s{3,}y\s{9,}\s*$", table) is None


@server_config
def test_cmd_obj_root_node(manager):
    base = f"-s {manager.sockfile} -f ok"
    cmd_no_root = base
    cmd_with_root = f"{base} -o root"

    assert run_qtile_cmd(cmd_no_root, no_eval=True) == run_qtile_cmd(cmd_with_root, no_eval=True)
